# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Benoît Pauwels - refactoring
"""
Updates a trust parameter according to a decreases ratio
********************************************************
"""

from __future__ import division, unicode_literals

import logging

from numpy import divide, maximum, minimum, multiply

LOGGER = logging.getLogger(__name__)


class TrustUpdater(object):
    """Updates the trust parameter."""

    def __init__(self, thresholds=None, multipliers=None, bound=None):
        """Initializer.

        :param thresholds: thresholds for the decreases ratio
        :type thresholds: tuple
        :param multipliers: multipliers for the trust parameter
        :type multipliers: tuple
        :param bound: (lower or upper) bound for the trust parameter
        """
        if not isinstance(thresholds, tuple):
            raise ValueError(
                "The thresholds must be input as a tuple; "
                "input of type " + type(thresholds) + " was provided."
            )
        self._ratio_thresholds = thresholds
        if not isinstance(multipliers, tuple):
            raise ValueError(
                "The multipliers must be input as a tuple; "
                "input of type " + type(multipliers) + " was provided."
            )
        self._param_multipliers = multipliers
        self._param_bound = bound  # bound for the trust parameter

    def _check(self):
        """Checks attributes consistency."""
        raise NotImplementedError()

    def update(self, ratio, parameter):
        """Updates the trust parameter relative to the decreases ratio value. Method to
        be overidden by subclasses.

        :param ratio: decreases ratio
        :param parameter: trust parameter (radius or penalty)
        :returns: new trust parameter, iteration success
        :rtype: bool
        """
        raise NotImplementedError()


class PenaltyUpdater(TrustUpdater):
    """Updates the penalty parameter."""

    def __init__(self, thresholds=(0.0, 0.25), multipliers=(0.5, 2.0), bound=1e-6):
        """Initializer.

        :param thresholds: thresholds for the decreases ratio
        :type thresholds: tuple
        :param multipliers: multipliers for the penalty parameter
        :type multipliers: tuple
        :param bound: lower bound for the penalty parameter
        """
        super(PenaltyUpdater, self).__init__(thresholds, multipliers, bound)
        self._check()

    def _check(self):
        """Checks the attributes of the penalty updater."""
        # Check the thresholds:
        if len(self._ratio_thresholds) != 2:
            raise ValueError(
                "There must be exactly two thresholds for the "
                "decreases ratio; " + str(len(self._ratio_thresholds)) + " were given."
            )
        update_thresh = self._ratio_thresholds[0]
        nonexp_thresh = self._ratio_thresholds[1]
        if update_thresh > nonexp_thresh:
            raise ValueError(
                "The update threshold ("
                + str(update_thresh)
                + ") must be lower than or equal to "
                + "the non-expansion threshold ("
                + str(nonexp_thresh)
                + ")."
            )
        # Check the multipliers:
        if len(self._param_multipliers) != 2:
            raise ValueError(
                "There must be exactly two multipliers for the "
                "penalty parameter; "
                + str(len(self._ratio_thresholds))
                + " were given."
            )
        contract_fact = self._param_multipliers[0]
        expan_fact = self._param_multipliers[1]
        if contract_fact >= 1.0:
            raise ValueError(
                "The contraction factor ("
                + str(contract_fact)
                + ") must be lower than one."
            )
        if expan_fact < 1.0:
            raise ValueError(
                "The expansion factor ("
                + str(expan_fact)
                + ") must be greater than or equal to one."
            )

    def update(self, ratio, parameter):
        """Updates the penalty parameter.

        :param ratio: decreases ratio
        :param parameter: penalty parameter
        :returns: new penalty parameter, iteration success (boolean)
        """
        # The iteration is declared successful if and only if the ratio is
        #         greater than or equal to the lower threshold.
        success = ratio >= self._ratio_thresholds[0]
        # If the ratio is greater than or equal to the upper threshold, the
        # penalty parameter is not increased, otherwise it is increased.
        if ratio >= self._ratio_thresholds[1]:
            new_penalty = parameter * self._param_multipliers[0]
            if self._param_bound is not None and new_penalty < self._param_bound:
                new_penalty = 0.0
        else:
            if self._param_bound is not None and parameter == 0.0:
                new_penalty = self._param_bound
            else:
                new_penalty = parameter * self._param_multipliers[1]
        return new_penalty, success


class RadiusUpdater(TrustUpdater):
    """Updates the trust region radius."""

    def __init__(self, thresholds=(0.0, 0.25), multipliers=(0.5, 2.0), bound=1000.0):
        """Initializer.

        :param thresholds: thresholds for the decreases ratio
        :type thresholds: tuple
        :param multipliers: multipliers for the region radius
        :type multipliers: tuple
        :param bound: lower bound for the region radius
        """
        super(RadiusUpdater, self).__init__(thresholds, multipliers, bound)
        self._check()

    def _check(self):
        """Checks the attributes of the radius updater."""
        # Check the thresholds:
        if len(self._ratio_thresholds) != 2:
            raise ValueError(
                "There must be exactly two thresholds for the "
                "decreases ratio; " + str(len(self._ratio_thresholds)) + " were given."
            )
        update_thresh = self._ratio_thresholds[0]
        noncontract_thresh = self._ratio_thresholds[1]
        if update_thresh > noncontract_thresh:
            raise ValueError(
                "The update threshold ("
                + str(update_thresh)
                + ") must be lower than or equal to the "
                "non-contraction threshold (" + str(noncontract_thresh) + ")."
            )
        # Check the multipliers:
        if len(self._param_multipliers) != 2:
            raise ValueError(
                "There must be exactly two multipliers for the "
                "region radius; " + str(len(self._ratio_thresholds)) + " were  given."
            )
        contract_fact = self._param_multipliers[0]
        expan_fact = self._param_multipliers[1]
        if contract_fact > 1.0:
            raise ValueError(
                "The contraction factor ("
                + str(contract_fact)
                + ") must be lower than or equal to one."
            )
        if expan_fact <= 1.0:
            raise ValueError(
                "The expansion factor ("
                + str(expan_fact)
                + ") must be greater than one."
            )

    def update(self, ratio, parameter):
        """Updates the trust radius.

        :param ratio: decreases ratio
        :param parameter: region radius
        :returns: new region radius, iteration success (boolean)
        """
        # The iteration is declared successful if and only if the ratio is
        # greater than or equal to the lower threshold.
        success = ratio >= self._ratio_thresholds[0]
        # If the ratio is greater than or equal to the upper threshold, the
        # region radius is not decreased, otherwise it is decreased.
        if ratio >= self._ratio_thresholds[1]:
            new_radius = parameter * self._param_multipliers[1]
            if self._param_bound is not None:
                new_radius = min(new_radius, self._param_bound)
        else:
            new_radius = parameter * self._param_multipliers[0]
        return new_radius, success


class BoundsUpdater(object):
    """Updates trust bounds, i.e. trust ball w.r.t.

    the infinity norm.
    """

    def __init__(self, lower_bounds, upper_bounds, normalize=False):
        """Initializer.

        :param lower_bounds: reference lower bounds
        :type lower_bounds: ndarray
        :param upper_bounds: reference upper bounds
        :type upper_bounds: ndarray
        :param normalize: if True the radius is applied to the normalized
            bounds
        :param normalize: bool, optional
        """
        self._lower_bounds = lower_bounds
        self._upper_bounds = upper_bounds
        self._normalized_update = normalize

    def update(self, radius, center):
        """Updates the trust bounds.

        :param radius: region radius w.r.t. the infinity norm
        :type radius: float
        :param center: region center
        :type center: ndarray
        :returns: new region radius, iteration success (boolean)
        """
        if not self._normalized_update:
            low_bnds, upp_bnds = self._compute_trust_bounds(
                self._lower_bounds, self._upper_bounds, center, radius
            )
        else:
            # Compute the normalized coordinate-specific radii:
            # radius_i = radius * 0.5 * (upper_bound_i - lower_bound_i)
            radii = radius * 0.5 * (self._upper_bounds - self._lower_bounds)
            # Compute the trust bounds
            low_bnds, upp_bnds = self._compute_trust_bounds(
                self._lower_bounds, self._upper_bounds, center, radii
            )

        return low_bnds, upp_bnds

    @staticmethod
    def _compute_trust_bounds(lower_bounds, upper_bounds, center, radius):
        """Updates bounds based on a ball center and ball radius w.r.t. the infinity
        norm.

        :param lower_bounds: lower bounds to be updated
        :type lower_bounds: ndarray
        :param upper_bounds: upper bounds to be updated
        :type upper_bounds: ndarray
        :param center: ball center
        :type center: ndarray
        :param radius: ball radius (same for all coordinate or coordinate-specific)
        :type radius: float or ndarray
        :returns: updated lower bounds, updated upper bounds
        :rtype: ndarray, ndarray
        """
        lower_trust_bounds = minimum(
            maximum(center - radius, lower_bounds), upper_bounds
        )
        upper_trust_bounds = maximum(
            minimum(center + radius, upper_bounds), lower_bounds
        )
        return lower_trust_bounds, upper_trust_bounds

    def _normalize(self, x_vect):
        """Normalize a vector coordinates to [-1, 1]."""
        x_norm = 2.0 * x_vect - self._upper_bounds - self._lower_bounds
        x_norm = divide(x_norm, self._upper_bounds - self._lower_bounds)
        return x_norm

    def _unnormalize(self, x_norm):
        """Unnormalize a vector coordinates from [-1, 1]."""
        x_vect = multiply(x_norm, (self._upper_bounds - self._lower_bounds))
        x_vect = (x_vect + self._upper_bounds + self._lower_bounds) / 2.0
        return x_vect



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Francois Gallard
"""Design of experiments from custom data."""
from __future__ import division, unicode_literals

import logging
from typing import Dict, List, Optional, Sequence, TextIO, Union

from numpy import apply_along_axis, atleast_2d, loadtxt, ndarray

from gemseo.algos.doe.doe_lib import DOELibrary
from gemseo.utils.py23_compat import Path

OptionType = Optional[Union[str, int, float, bool, List[str], Path, TextIO, ndarray]]

LOGGER = logging.getLogger(__name__)


class CustomDOE(DOELibrary):
    """A design of experiments from samples provided as a file or an array.

    The samples are provided
    either as a file in text or csv format
    or as a sequence of sequences of numbers,
    e.g. a 2D numpy array.

    A csv file format is assumed to have a header
    whereas a text file (extension .txt) does not.
    """

    ALGO_LIST = ["CustomDOE"]
    COMMENTS_KEYWORD = "comments"
    DELIMITER_KEYWORD = "delimiter"
    SKIPROWS_KEYWORD = "skiprows"
    DOE_FILE = "doe_file"
    SAMPLES = "samples"

    def __init__(self):  # type: (...) -> None
        super(CustomDOE, self).__init__()
        self.file_dv_names_list = None

        desc = {
            "CustomDOE": (
                "This samples are provided "
                "either as a file in text or csv format "
                "or as a sequence of sequences of numbers."
            )
        }
        for algo in self.ALGO_LIST:
            self.lib_dict[algo] = {
                DOELibrary.LIB: self.__class__.__name__,
                DOELibrary.INTERNAL_NAME: algo,
                DOELibrary.DESCRIPTION: desc[algo],
            }

    def _get_options(
        self,
        doe_file=None,  # type: Optional[Union[str, Path, TextIO]]
        samples=None,  # type: Optional[ndarray]
        delimiter=",",  # type: Optional[str]
        comments="#",  # type: Optional[Union[str,Sequence[str]]]
        skiprows=0,  # type: int
        max_time=0,  # type: float
        eval_jac=False,  # type: bool
        n_processes=1,  # type: int
        wait_time_between_samples=0.0,  # type: float
        **kwargs
    ):  # type: (...) -> Dict[str,OptionType]
        """Sets the options.

        :param doe_file: Either the file, the filename, or the generator to read.
        :type doe_file: str
        :param samples: The samples.
        :type samples: array
        :param delimiter: The string used to separate values.
            If None, use whitespace.
        :type delimiter: str
        :param comments:  The characters or list of characters
            used to indicate the start of a comment.
            None implies no comments.
        :type comments: str
        :param skiprows: skip the first `skiprows` lines
        :type skiprows: int
        :param eval_jac: evaluate jacobian
        :type eval_jac: bool
        :param n_processes: number of processes
        :type n_processes: int
        :param wait_time_between_samples: waiting time between two samples
        :type wait_time_between_samples: float
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param kwargs: additional arguments
        """
        wtbs = wait_time_between_samples
        return self._process_options(
            max_time=max_time,
            doe_file=doe_file,
            samples=samples,
            delimiter=delimiter,
            comments=comments,
            skiprows=skiprows,
            eval_jac=eval_jac,
            n_processes=n_processes,
            wait_time_between_samples=wtbs,
            **kwargs
        )

    def read_file(
        self,
        doe_file,  # type: Union[str, Path, TextIO]
        delimiter=",",  # type: Optional[str]
        comments="#",  # type: Optional[Union[str,Sequence[str]]]
        skiprows=0,  # type: int
    ):  # type: (...) -> ndarray
        """Read a file containing several samples (one per line) and return them.

        Args:
            doe_file: Either the file, the filename, or the generator to read.
            delimiter: The string used to separate values.
                If None, use whitespace.
            comments:  The characters or list of characters
                used to indicate the start of a comment.
                None implies no comments.
            skiprows: Skip the first `skiprows` lines.

        Returns:
            The samples.
        """
        try:
            samples = loadtxt(
                doe_file,
                comments=comments,
                delimiter=delimiter,
                skiprows=skiprows,
                unpack=False,
            )
            samples = atleast_2d(samples)
            if (
                samples.shape[1] != self.problem.dimension
                and self.problem.dimension == 1
            ):
                samples = samples.T
        except ValueError:
            LOGGER.error("Failed to load DOE input file: %s", doe_file)
            raise

        return samples

    def _generate_samples(
        self, **options  # type: OptionType
    ):  # type: (...) -> ndarray
        """
        Raises:
            ValueError: If the dimension is different from the problem one.
        """
        error_message = (
            "The algorithm CustomDOE requires "
            "either 'doe_file' or 'samples' as option"
        )
        samples = options.get(self.SAMPLES)
        if samples is None:
            doe_file = options.get(self.DOE_FILE)
            if doe_file is None:
                raise ValueError(error_message)
            samples = self.read_file(
                doe_file,
                comments=options[self.COMMENTS_KEYWORD],
                delimiter=options[self.DELIMITER_KEYWORD],
                skiprows=options[self.SKIPROWS_KEYWORD],
            )
        else:
            if options.get(self.DOE_FILE) is not None:
                raise ValueError(error_message)

        if samples.shape[1] != self.problem.dimension:
            raise ValueError(
                "Dimension mismatch between the problem ({}) and "
                " the samples ({})".format(self.problem.dimension, samples.shape[1])
            )

        samples = apply_along_axis(
            self.problem.design_space.normalize_vect, axis=1, arr=samples
        )

        return samples



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to create a joint probability distribution from the SciPy library.

The :class:`.SPComposedDistribution` class is a concrete class
inheriting from :class:`.ComposedDistribution` which is an abstract one.
SP stands for `scipy <https://docs.scipy.org/doc/scipy/reference/
tutorial/stats.html>`_ which is the library it relies on.

This class inherits from :class:`.SPDistribution`.
It builds a composed probability distribution
related to given random variables from a list of :class:`.SPDistribution` objects
implementing the probability distributions of these variables
based on the SciPy library and from a copula name.

.. note::

   A copula is a mathematical function used to define the dependence
   between random variables from their cumulative density functions.
   `See more <https://en.wikipedia.org/wiki/Copula_(probability_theory)>`__.
"""

from __future__ import division, unicode_literals

from typing import TYPE_CHECKING, Callable, Iterable, Sequence

if TYPE_CHECKING:
    from gemseo.uncertainty.distributions.scipy.distribution import SPDistribution

from numpy import array, ndarray

from gemseo.uncertainty.distributions.composed import ComposedDistribution


class SPComposedDistribution(ComposedDistribution):
    """Scipy composed distribution."""

    _COPULA = {ComposedDistribution._INDEPENDENT_COPULA: None}
    AVAILABLE_COPULA_MODELS = sorted(_COPULA.keys())

    def __init__(
        self,
        distributions,  # type: Sequence[SPDistribution]
        copula=ComposedDistribution._INDEPENDENT_COPULA,  # type: str
    ):  # type: (...) -> None # noqa: D205,D212,D415
        """
        Args:
            distributions (list(SPDistribution)): The distributions.
            copula (str, optional): A name of copula.
        """
        super(SPComposedDistribution, self).__init__(distributions, copula)
        self.distribution = distributions
        self._mapping = {}
        index = 0
        for marginal_index, marginal in enumerate(self.distribution):
            for submarginal_index in range(marginal.dimension):
                self._mapping[index] = (marginal_index, submarginal_index)
                index += 1
        self._set_bounds(distributions)

    def compute_cdf(
        self,
        vector,  # type: Iterable[float]
    ):  # noqa: D102
        # type: (...) -> ndarray
        tmp = []
        for index, value in enumerate(vector):
            id1 = self._mapping[index][0]
            id2 = self._mapping[index][1]
            tmp.append(self.distribution[id1].marginals[id2].cdf(value))
        return array(tmp)

    def compute_inverse_cdf(
        self,
        vector,  # type: Iterable[float]
    ):  # noqa: D102
        # type: (...) -> ndarray
        tmp = []
        for index, value in enumerate(vector):
            id1 = self._mapping[index][0]
            id2 = self._mapping[index][1]
            tmp.append(self.distribution[id1].marginals[id2].ppf(value))
        return array(tmp)

    def _pdf(
        self,
        index,  # type: int
    ):  # noqa: D102
        # type: (...) -> Callable
        id1 = self._mapping[index][0]
        id2 = self._mapping[index][1]

        def pdf(
            point,  # type: float
        ):
            # type: (...) -> float
            """Probability Density Function (PDF).

            Args:
                point: An evaluation point.

            Returns:
                The PDF value at the evaluation point.
            """
            return self.distribution[id1].marginals[id2].pdf(point)

        return pdf

    def _cdf(
        self,
        index,  # type: int
    ):  # noqa: D102
        # type: (...) -> Callable
        id1 = self._mapping[index][0]
        id2 = self._mapping[index][1]

        def cdf(
            level,  # type: float
        ):
            # type: (...) -> float
            """Cumulative Density Function (CDF).

            Args:
                level: A probability level.

            Returns:
                The CDF value for the probability level.
            """
            return self.distribution[id1].marginals[id2].cdf(level)

        return cdf



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Caching module to avoid multiple evaluations of a discipline
************************************************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.cache import AbstractFullCache
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.locks import synchronized
from gemseo.utils.multi_processing import RLock

LOGGER = logging.getLogger(__name__)


class MemoryFullCache(AbstractFullCache):
    """Cache using memory to cache all data."""

    def __init__(self, tolerance=0.0, name=None, is_memory_shared=True):
        """Initialize a dictionary to cache data.

        Initialize cache tolerance.
        By default, don't use approximate cache.
        It is up to the user to choose to optimize CPU time with this or not
        could be something like 2 * finfo(float).eps

        Parameters
        ----------
        tolerance : float
            Tolerance that defines if two input vectors
            are equal and cached data shall be returned.
            If 0, no approximation is made. Default: 0.
        name : str
            Name of the cache.
        is_memory_shared : bool
            If True, a shared memory dict is used to store the data,
            which makes the cache compatible with multiprocessing.
            WARNING: if set to False, and multiple disciplines point to
            the same cache or the process is multiprocessed, there may
            be duplicate computations because the cache will not be
            shared among the processes.

        Examples
        --------
        >>> from gemseo.caches.memory_full_cache import MemoryFullCache
        >>> cache = MemoryFullCache()
        """
        self.__is_memory_shared = is_memory_shared
        super(MemoryFullCache, self).__init__(tolerance, name)
        self.__init_data()

    def __init_data(self):
        """Initializes the local dict that stores the data.

        Either a shared memory dict or a basic dict.
        """
        if self.__is_memory_shared:
            self._data = self._manager.dict()
        else:
            self._data = {}

    def _duplicate_from_scratch(self):
        return MemoryFullCache(self.tolerance, self.name, self.__is_memory_shared)

    def _initialize_entry(self, sample_id):
        """Initialize an entry of the dataset if needed.

        :param int sample_id: sample ID.
        """
        template = {}
        self._data[sample_id] = template

    def _set_lock(self):
        """Sets a lock for multithreading, either from an external object or internally
        by using RLock()."""
        return RLock()

    def _has_group(self, sample_id, var_group):
        """Checks if the dataset has the particular variables group filled in.

        :param int sample_id: sample ID.
        :param str var_group: name of the variables group.
        :return: True if the variables group is filled in.
        :rtype: bool
        """
        return var_group in self._data.get(sample_id)

    @synchronized
    def clear(self):
        """Clear the cache.

        Examples
        --------
        >>> from gemseo.caches.memory_full_cache import MemoryFullCache
        >>> from numpy import array
        >>> cache = MemoryFullCache()
        >>> for index in range(5):
        >>>     data = {'x': array([1.])*index, 'y': array([.2])*index}
        >>>     cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_length()
        5
        >>> cache.clear()
        >>> cache.get_length()
        0
        """
        super(MemoryFullCache, self).clear()
        self.__init_data()

    def _read_data(self, group_number, group_name):
        """Read a data dict in the hdf.

        :param group_name: name of the group where data is written
        :param group_number: number of the group
        :returns: data dict and jacobian
        """
        result = self._data[group_number].get(group_name)
        if group_name == self.JACOBIAN_GROUP and result is not None:
            result = DataConversion.dict_to_jac_dict(result)
        return result

    def _write_data(self, values, names, var_group, sample_id):
        """Writes data associated with a variables group and a sample ID into the
        dataset.

        :param dict values: data dictionary where keys are variables names
            and values are variables values (numpy arrays).
        :param list(str) names: list of input data names to write.
        :param str var_group: name of the variables group,
            either AbstractCache.INPUTS_GROUP, AbstractCache.OUTPUTS_GROUP or
            AbstractCache.JACOBIAN_GROUP.
        :param int sample_id: sample ID.
        """
        data = self._data[sample_id]
        data[var_group] = {name: values[name] for name in names}
        self._data[sample_id] = data

    @property
    def copy(self):
        """Copy cache."""
        cache = self._duplicate_from_scratch()
        cache.merge(self)
        return cache



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                  initial documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable discipline
===================

The :mod:`~gemseo.problems.scalable.data_driven.discipline`
implements the concept of scalable discipline.
This is a particular discipline
built from a input-output learning dataset associated with a function
and generalizing its behavior to a new user-defined problem dimension,
that is to say new user-defined input and output dimensions.

Alone or in interaction with other objects of the same type,
a scalable discipline can be used to compare the efficiency of an algorithm
applying to disciplines with respect to the problem dimension,
e.g. optimization algorithm, surrogate model, MDO formulation, MDA, ...

The :class:`.ScalableDiscipline` class implements this concept.
It inherits from the :class:`.MDODiscipline` class
in such a way that it can easily be used in a :class:`.Scenario`.
It is composed of a :class:`.ScalableModel`.


The user only needs to provide:

- the name of a class overloading :class:`.ScalableModel`,
- a dataset as an :class:`.Dataset`
- variables sizes as a dictionary
  whose keys are the names of inputs and outputs
  and values are their new sizes.
  If a variable is missing, its original size is considered.

The :class:`.ScalableModel` parameters can also be filled in,
otherwise the model uses default values.
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.discipline import MDODiscipline
from gemseo.problems.scalable.data_driven.factory import ScalableModelFactory
from gemseo.utils.data_conversion import DataConversion

LOGGER = logging.getLogger(__name__)


class ScalableDiscipline(MDODiscipline):
    """Scalable discipline."""

    def __init__(self, name, data, sizes=None, **parameters):
        """Constructor.

        :param str name: scalable model class name.
        :param Dataset data: learning dataset.
        :param dict sizes: sizes of input and output variables.
            If None, use the original sizes.
            Default: None.
        :param parameters: model parameters
        """
        create = ScalableModelFactory().create
        self.scalable_model = create(name, data=data, sizes=sizes, **parameters)
        super(ScalableDiscipline, self).__init__(self.scalable_model.name)
        self.initialize_grammars(data)
        self._default_inputs = self.scalable_model.default_inputs
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        self.add_differentiated_inputs(self.get_input_data_names())
        self.add_differentiated_outputs(self.get_output_data_names())

    def initialize_grammars(self, data):
        """Initialize input and output grammars from data names.

        :param Dataset data: learning dataset.
        """
        self.input_grammar.initialize_from_data_names(data.get_names(data.INPUT_GROUP))
        self.output_grammar.initialize_from_data_names(
            data.get_names(data.OUTPUT_GROUP)
        )

    def _run(self):
        """Runs the scalable discipline and stores the output values."""
        output_value = self.scalable_model.scalable_function(self.local_data)
        self.local_data.update(output_value)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the Jacobian of outputs wrt inputs and store the values.

        :param inputs: list of input variables. Default value: None.
        :type inputs: list(str)
        :param outputs: list of output functions.  Default value: None.
        :type outputs: list(str)
        """
        self._init_jacobian(inputs, outputs, with_zeros=True)
        jac = self.scalable_model.scalable_derivatives(self.local_data)
        inputs_names = self.scalable_model.inputs_names
        jac = {
            fname: DataConversion.array_to_dict(
                jac[fname], inputs_names, self.scalable_model.sizes
            )
            for fname in self.get_output_data_names()
        }
        self.jac = jac



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Setup entry point."""

import setuptools

import versioneer

setuptools.setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: François Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Dummy linear discipline generator
=================================

A utility that generates dummy disciplines from a specification.
The inputs and output names are specified by the user.
A linear random dependency between the inputs and outputs is created.
The size of the inputs and outputs can be parametrized by the user.
The MDA of the generated disciplines will always converge because all the outputs
are in [0, 1] if the inputs are in [0, 1].
The analytic Jacobian is provided.
"""

import string
from itertools import islice, permutations
from typing import List

from numpy import arange, array
from numpy.random import shuffle

from gemseo.core.discipline import MDODiscipline
from gemseo.problems.scalable.linear.linear_discipline import LinearDiscipline

DESC_5_DISC = [
    ("A", ["b"], ["a", "c"]),
    ("B", ["a"], ["b"]),
    ("C", ["c", "e"], ["d"]),
    ("D", ["d"], ["e", "f"]),
    ("E", ["f"], []),
]

DESC_16_DISC = [
    ("A", ["a"], ["b"]),
    ("B", ["c"], ["a", "n"]),
    ("C", ["b", "d"], ["c", "e"]),
    ("D", ["f"], ["d", "g"]),
    ("E", ["e"], ["f", "h", "o"]),
    ("F", ["g", "j"], ["i"]),
    ("G", ["i", "h"], ["k", "l"]),
    ("H", ["k", "m"], ["j"]),
    ("I", ["l"], ["m", "w"]),
    ("J", ["n", "o"], ["p", "q"]),
    ("K", ["y"], ["x"]),
    ("L", ["w", "x"], ["y", "z"]),
    ("M", ["p", "s"], ["r"]),
    ("N", ["r"], ["t", "u"]),
    ("O", ["q", "t"], ["s", "v"]),
    ("P", ["u", "v", "z"], ["obj"]),
]

DESC_3_DISC_WEAK = [
    ("A", ["x"], ["a"]),
    ("B", ["x", "a"], ["b"]),
    ("C", ["x", "a"], ["c"]),
]

DESC_4_DISC_WEAK = [
    ("A", ["x"], ["a"]),
    ("B", ["x", "a"], ["b"]),
    ("C", ["x", "a"], ["c"]),
    ("D", ["b", "c"], ["d"]),
]

DESC_DISC_REPEATED = [
    ("A", ["a"], ["b"]),
    ("A", ["a"], ["b"]),
    ("A", ["c"], ["d"]),
]

LETTERS = array(list(string.ascii_uppercase))


def _get_disc_names(
    nb_of_names,  # type: int
):  # type: (...) -> List[str]
    """Generate names from alphabet characters combinations.

    For a given number of names, generates combinations of the characters.

    Args:
        nb_of_names: The number of names to generate.

    Returns:
        The names.
    """
    n_letters = 1

    while len(LETTERS) ** n_letters < nb_of_names:
        n_letters += 1

    return ["".join(c) for c in islice(permutations(LETTERS, n_letters), nb_of_names)]


def create_disciplines_from_sizes(
    nb_of_disc,  # type: int
    nb_of_total_disc_io,  # type: int
    nb_of_disc_inputs=1,  # type: int
    nb_of_disc_outputs=1,  # type: int
    inputs_size=1,  # type: int
    outputs_size=1,  # type: int
    grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,  # type: str
):  # type: (...) -> List[LinearDiscipline]
    """Generate :class:`.LinearDiscipline`s according to a specification.

    The names of the disciplines are generated from capital letters words.
    The names of the inputs and outputs are generated from string representations of
    integers.

    Args:
        nb_of_disc: The total number of disciplines.
        nb_of_total_disc_io: The total number of input and output data names
            in the overall process.
        nb_of_disc_inputs: The number of disciplines inputs, same for all
            disciplines.
        nb_of_disc_outputs: The number of disciplines outputs, same for all
            disciplines.
        inputs_size: The size of the input vectors,
            each input data is of shape (inputs_size,).
        outputs_size: The size of the output vectors,
            each output data is of shape (outputs_size,).
        grammar_type: The type of grammars used by the discipline.

    Returns:
        The :class:`.LinearDiscipline`s.

    Raises:
        ValueError: If the number of disciplines is inconsistent with the
            total number of inputs or outputs.
    """
    if nb_of_disc_inputs > nb_of_total_disc_io:
        raise ValueError(
            "The number of disciplines inputs must be lower "
            "or equal than the total number of disciplines io"
        )

    if nb_of_disc_outputs > nb_of_total_disc_io:
        raise ValueError(
            "The number of disciplines outputs must be lower "
            "or equal than the total number of disciplines io"
        )

    disc_names = _get_disc_names(nb_of_disc)

    disc_descriptions = []

    for disc_name in disc_names:
        # Choose inputs among all io
        shuff_names = arange(nb_of_total_disc_io)
        shuffle(shuff_names)
        in_names = array(shuff_names[:nb_of_disc_inputs], dtype="str").tolist()

        shuff_names = arange(nb_of_total_disc_io)
        shuffle(shuff_names)
        out_names = array(shuff_names[:nb_of_disc_outputs], dtype="str").tolist()

        disc_descriptions.append((disc_name, in_names, out_names))

    return create_disciplines_from_desc(
        disc_descriptions,
        inputs_size=inputs_size,
        outputs_size=outputs_size,
        grammar_type=grammar_type,
    )


def create_disciplines_from_desc(
    disc_descriptions,  # Sequence[Tuple[str,Sequence[str],Sequence[str]]]
    inputs_size=1,  # type: int
    outputs_size=1,  # type: int
    grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,  # type: str
):  # type: (...) -> List[LinearDiscipline]
    """Generate :class:`.LinearDiscipline`s according to a specification.

    The specification is as follows:

    .. code-block:: python

        [
        ("Disc_name1", ["in1"], ["out1", "out2"]),
        ("Disc_name2", ["in2", "out1"], ["out3", "out2"]),
        ]

    This will generate two disciplines:
      - One named "Disc_name1" with the inputs ["in1"] and the outputs ["out1", "out2"].
      - Another named "Disc_name2" with the inputs ["in2", "out1"]
        and the outputs ["out3", "out2"].

    Args:
        disc_descriptions: The specification of the disciplines,
            each item is (name, inputs_names, outputs_names),
            disciplines names may be non unique.
        inputs_size: The size of the input vectors,
            each input data is of shape (inputs_size,).
        outputs_size: The size of the output vectors,
            each output data is of shape (outputs_size,).
        grammar_type: The type of grammars used by the disciplines.

    Returns:
        The :class:`.LinearDiscipline`s.
    """
    return [
        LinearDiscipline(
            name, input_names, output_names, inputs_size, outputs_size, grammar_type
        )
        for name, input_names, output_names in disc_descriptions
    ]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Provide base test class stub for testing bilevel also for |g| plugins."""
from __future__ import division, unicode_literals

import unittest
from copy import deepcopy

from gemseo.core.mdo_scenario import MDOScenario
from gemseo.problems.sobieski.wrappers import (
    SobieskiAerodynamics,
    SobieskiMission,
    SobieskiProblem,
    SobieskiPropulsion,
    SobieskiStructure,
)


def build_system_scenario(disciplines, **options):
    """

    :param disciplines:

    """
    # Maximize range (Breguet)
    ds = SobieskiProblem().read_design_space()
    # Add a coupling to DV but bielevel filters it
    sc_system = MDOScenario(
        disciplines,
        formulation="BiLevel",
        objective_name="y_4",
        design_space=ds.filter(["x_shared", "y_14"]),
        maximize_objective=True,
        **options
    )
    assert sc_system.formulation.opt_problem.design_space.variables_names == [
        "x_shared"
    ]
    sc_system.set_differentiation_method("finite_differences", step=1e-6)
    return sc_system


class TestBilevelFormulationBase(unittest.TestCase):
    """Helper class for testing bilevel formulations."""

    @classmethod
    def setUpClass(cls):
        cls.propulsion = SobieskiPropulsion()
        cls.aerodynamics = SobieskiAerodynamics()
        cls.struct = SobieskiStructure()
        cls.mission = SobieskiMission()

    def build_scenarios(self):
        """"""
        ds = SobieskiProblem().read_design_space()
        sc_prop = MDOScenario(
            disciplines=[self.propulsion],
            formulation="DisciplinaryOpt",
            objective_name="y_34",
            design_space=deepcopy(ds).filter("x_3"),
            name="PropulsionScenario",
        )

        # Maximize L/D
        sc_aero = MDOScenario(
            disciplines=[self.aerodynamics],
            formulation="DisciplinaryOpt",
            objective_name="y_24",
            design_space=deepcopy(ds).filter("x_2"),
            name="AerodynamicsScenario",
            maximize_objective=True,
        )

        # Maximize log(aircraft total weight / (aircraft total weight - fuel
        # weight))
        sc_str = MDOScenario(
            disciplines=[self.struct],
            formulation="DisciplinaryOpt",
            objective_name="y_11",
            design_space=deepcopy(ds).filter("x_1"),
            name="StructureScenario",
            maximize_objective=True,
        )

        return sc_str, sc_aero, sc_prop

    def build_bilevel(self, **options):
        """"""
        sub_scenarios = self.build_scenarios()
        sub_disciplines = list(sub_scenarios) + [self.mission]
        for sc in sub_scenarios:
            sc.default_inputs = {"max_iter": 5, "algo": "SLSQP"}
        system_scenario = build_system_scenario(sub_disciplines, **options)
        return system_scenario



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Machine learning API
--------------------

The machine learning API provides methods for creating new and loading
existing machine learning models. It also provides methods for listing
available models and options.
"""
from __future__ import division, unicode_literals

import logging
from typing import Dict, List, Optional, Union

from gemseo.api import _get_schema
from gemseo.core.dataset import Dataset
from gemseo.mlearning.classification.classification import MLClassificationAlgo
from gemseo.mlearning.cluster.cluster import MLClusteringAlgo
from gemseo.mlearning.core.ml_algo import MLAlgo, TransformerType
from gemseo.mlearning.core.supervised import MLSupervisedAlgo
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.mlearning.transform.scaler.min_max_scaler import MinMaxScaler

LOGGER = logging.getLogger(__name__)

# pylint: disable=import-outside-toplevel


def get_mlearning_models():  # type:(...) -> List[str]
    """Get available machine learning algorithms.

    Returns:
        The available machine learning algorithms.

    See also
    --------
    import_mlearning_model
    create_mlearning_model
    get_mlearning_options
    import_mlearning_model
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory()
    return factory.models


def get_regression_models():  # type:(...) -> List[str]
    """Get available regression models.

    Returns:
        The available regression models.

    See also
    --------
    create_regression_model
    get_regression_options
    import_regression_model
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory()
    return factory.models


def get_classification_models():  # type:(...) -> List[str]
    """Get available classification models.

    Returns:
        The available classification models.

    See also
    --------
    create_classification_model
    get_classification_options
    import_classification_model
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory()
    return factory.models


def get_clustering_models():  # type:(...) -> List[str]
    """Get available clustering models.

    Returns:
        The available clustering models.

    See also
    --------
    create_clustering_model
    get_clustering_options
    import_clustering_model
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory()
    return factory.models


def create_mlearning_model(
    name,  # type: str
    data,  # type: Dataset
    transformer=None,  # type: Optional[TransformerType]
    **parameters
):  # type:(...) -> MLAlgo
    """Create a machine learning algorithm from a learning dataset.

    Args:
        name: The name of the machine learning algorithm.
        data: The learning dataset.
        transformer: The strategies to transform the variables.
            Values are instances of :class:`.Transformer`
            while keys are names of either variables or groups of variables.
            If None, do not transform the variables.
        parameters: The parameters of the machine learning algorithm.

    Returns:
        A machine learning model.

    See also
    --------
    get_mlearning_models
    get_mlearning_options
    import_mlearning_model
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


minmax_inputs = {Dataset.INPUT_GROUP: MinMaxScaler()}


def create_regression_model(
    name,  # type: str
    data,  # type: Dataset
    transformer=MLRegressionAlgo.DEFAULT_TRANSFORMER,  # type:Optional[TransformerType]
    **parameters
):  # type: (...) -> MLRegressionAlgo
    """Create a regression model from a learning dataset.

    Args:
        name: The name of the regression algorithm.
        data: The learning dataset.
        transformer: The strategies to transform the variables.
            Values are instances of :class:`.Transformer`
            while keys are names of either variables or groups of variables.
            If None, do not transform the variables.
        parameters: The parameters of the regression model.

    Returns:
        A regression model.

    See also
    --------
    get_regression_models
    get_regression_options
    import_regression_model
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory()
    if (
        name == "PCERegression"
        and isinstance(transformer, dict)
        and Dataset.INPUT_GROUP in transformer
    ):
        LOGGER.warning(
            "Remove input data transformation because "
            "PCERegression does not support transformers."
        )
        del transformer[Dataset.INPUT_GROUP]
    return factory.create(name, data=data, transformer=transformer, **parameters)


def create_classification_model(
    name,  # type: str
    data,  # type: Dataset
    transformer=MLSupervisedAlgo.DEFAULT_TRANSFORMER,  # type:Optional[TransformerType]
    **parameters
):  # type: (...) -> MLClassificationAlgo
    """Create a classification model from a learning dataset.

    Args:
        name: The name of the classification algorithm.
        data: The learning dataset.
        transformer: The strategies to transform the variables.
            Values are instances of :class:`.Transformer`
            while keys are names of either variables or groups of variables.
            If None, do not transform the variables.
        parameters: The parameters of the classification model.

    Returns:
        A classification model.

    See also
    --------
    get_classification_models
    get_classification_options
    import_classification_model
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


def create_clustering_model(
    name,  # type: str
    data,  # type: Dataset
    transformer=None,  # type:Optional[TransformerType]
    **parameters
):  # type: (...) -> MLClusteringAlgo
    """Create a clustering model from a learning dataset.

    Args:
        name: The name of the clustering algorithm.
        data: The learning dataset.
        transformer: The strategies to transform the variables.
            Values are instances of :class:`.Transformer`
            while keys are names of either variables or groups of variables.
            If None, do not transform the variables.
        parameters: The parameters of the clustering model.

    Returns:
        A clustering model.

    See also
    --------
    get_clustering_models
    get_clustering_options
    import_clustering_model
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


def import_mlearning_model(
    directory,  # type: str
):  # type: (...) -> MLAlgo
    """Import a machine learning algorithm from a directory.

    Args:
        directory: The name of the directory.

    Returns:
        A machine learning model.

    See also
    --------
    create_mlearning_model
    get_mlearning_models
    get_mlearning_options
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory()
    return factory.load(directory)


def import_regression_model(
    directory,  # type: str
):  # type: (...) -> MLRegressionAlgo
    """Import a regression model from a directory.

    Args:
        directory: The name of the directory.

    Returns:
        A regression model.

    See also
    --------
    create_regression_model
    get_regression_models
    get_regression_options
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory()
    return factory.load(directory)


def import_classification_model(
    directory,  # type: str
):  # type: (...) -> MLClassificationAlgo
    """Import a classification model from a directory.

    Args:
        directory: The name of the directory.

    Returns:
        A classification model.

    See also
    --------
    create_classification_model
    get_classification_models
    get_classification_options
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory()
    return factory.load(directory)


def import_clustering_model(
    directory,  # type: str
):  # type: (...) -> MLClusteringAlgo
    """Import a clustering model from a directory.

    Args:
        directory: The name of the directory.

    Returns:
        A clustering model.

    See also
    --------
    create_clustering_model
    get_clustering_models
    get_clustering_options
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory()
    return factory.load(directory)


def get_mlearning_options(
    model_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=True,  # type:bool
):  # type: (...) -> Union[Dict[str,str],str]
    """Find the available options for a machine learning algorithm.

    Args:
        model_name: The name of the machine learning algorithm.
        output_json: Apply JSON format for the schema.
        pretty_print: Whether to print the schema in a pretty table.

    Returns:
        The options schema of the machine learning algorithm.

    See also
    --------
    create_mlearning_model
    get_mlearning_models
    import_mlearning_model
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_regression_options(
    model_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=True,  # type:bool
):  # type: (...) -> Union[Dict[str,str],str]
    """Find the available options for a regression model.

    Args:
        model_name: The name of the regression model.
        output_json: Apply JSON format for the schema.
        pretty_print: Print the schema in a pretty table.

    Returns:
        The options schema of the regression model.

    See also
    --------
    create_regression_model
    get_regression_models
    import_regression_model
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_classification_options(
    model_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=True,  # type:bool
):  # type: (...) -> Union[Dict[str,str],str]
    """Find the available options for a classification model.

    Args:
        model_name: The name of the classification model.
        output_json: Apply JSON format for the schema.
        pretty_print: Print the schema in a pretty table.

    Returns:
        The options schema of the classification model.

    See also
    --------
    create_classification_model
    get_classification_models
    import_classification_model
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_clustering_options(
    model_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=True,  # type:bool
):  # type: (...) -> Union[Dict[str,str],str]
    """Find the available options for clustering model.

    Args:
        model_name: The name of the clustering model.
        output_json: Apply JSON format for the schema.
        pretty_print: Print the schema in a pretty table.

    Returns:
        The options schema of the clustering model.

    See also
    --------
    create_clustering_model
    get_clustering_models
    import_clustering_model
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                  initial documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Scalable model factory
======================

This module contains the :class:`.ScalableModelFactory` which is a factory
to create a :class:`.ScalableModel` from its class name by means of the
:meth:`.ScalableModelFactory.create` method. It is also possible to get a list
of available scalable models
(see :attr:`.ScalableModelFactory.scalable_models` method)
and to check is a type of scalable model is available
(see :meth:`.ScalableModelFactory.is_available` method)
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.factory import Factory
from gemseo.problems.scalable.data_driven.model import ScalableModel

LOGGER = logging.getLogger(__name__)


class ScalableModelFactory(object):
    """This factory instantiates a class:`.ScalableModel` from its class name.

    The class can be internal to |g| or located in an external module whose path is
    provided to the constructor.
    """

    def __init__(self):
        """Initializes the factory: scans the directories to search for subclasses of
        ScalableModel.

        Searches in "GEMSEO_PATH" and gemseo.caches
        """
        self.factory = Factory(ScalableModel, ("gemseo.problems.scalable",))

    def create(self, model_name, data, sizes=None, **parameters):
        """Create a scalable model.

        :param str model_name: name of the scalable model (its classname)
        :param Dataset data: learning dataset.
        :param dict sizes: sizes of input and output variables.
            If None, use the original sizes.
            Default: None.
        :param parameters: model parameters
        :return: model_name scalable model
        """
        return self.factory.create(model_name, data=data, sizes=sizes, **parameters)

    @property
    def scalable_models(self):
        """Lists the available classes for scalable models.

        :returns: the list of classes names.
        :rtype: list(str)
        """
        return self.factory.classes

    def is_available(self, model_name):
        """Checks the availability of a scalable model.

        :param str model_name:  model_name of the scalable model.
        :returns: True if the scalable model is available.
        :rtype: bool
        """
        return self.factory.is_available(model_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable problem - Variables
****************************
"""
from __future__ import division, unicode_literals

X_LOCAL_NAME_BASIS = "x_local"
U_LOCAL_NAME_BASIS = "u_local"


def get_u_local_name(index):
    """Returns the name of the local uncertain parameter associated with an index.

    :param int index: index.
    """
    return "{}_{}".format(U_LOCAL_NAME_BASIS, index)


def get_x_local_name(index):
    """Returns the name of the local design parameter associated with an index.

    :param int index: index.
    """
    return "{}_{}".format(X_LOCAL_NAME_BASIS, index)


X_SHARED_NAME = "x_shared"


def get_coupling_name(index):
    """Returns the name of the coupling variable associated with an index.

    :param int index: index.
    """
    return "y_{}".format(index)


def get_constraint_name(index):
    """Returns the name of the constraint associated with an index.

    :param int index: index.
    """
    return "cstr_{}".format(index)


OBJECTIVE_NAME = "obj"


def check_consistency(n_shared, n_local, n_coupling):
    """Check if n_shared is an integer and if n_local and n_coupling are list of
    integers with the same length."""
    if not isinstance(n_shared, int) or n_shared < 1:
        raise TypeError("n_shared must be an integer > 0.")
    if not isinstance(n_local, list):
        raise TypeError("n_local must be a list of integers > 0.")
    if not isinstance(n_coupling, list):
        raise TypeError("n_coupling must be a list of integers > 0.")
    if not all([isinstance(val, int) for val in n_local]):
        raise TypeError("n_local must be a list of integers > 0.")
    if not all([val > 0 for val in n_local]):
        raise TypeError("n_local must be a list of integers > 0.")
    if not all([isinstance(val, int) for val in n_coupling]):
        raise TypeError("n_coupling must be a list of integers > 0.")
    if not all([val > 0 for val in n_coupling]):
        raise TypeError("n_coupling must be a list of integers > 0.")
    if len(n_local) != len(n_coupling):
        raise ValueError(
            "n_local and n_coupling must have the same length "
            "which is equal to the number of disciplines"
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                        initial documentation
#        :author: Francois Gallard, Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The Individual Discipline Feasible (IDF) formulation."""
from __future__ import division, unicode_literals

import logging
from typing import Iterable, List, Sequence, Tuple

from numpy import abs as np_abs
from numpy import concatenate, eye, ndarray, ones_like, zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.core.chain import MDOParallelChain
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import ExecutionSequence, ExecutionSequenceFactory
from gemseo.core.formulation import MDOFormulation
from gemseo.core.function import MDOFunction
from gemseo.mda.mda_chain import MDAChain

LOGGER = logging.getLogger(__name__)


class IDF(MDOFormulation):
    """The Individual Discipline Feasible (IDF) formulation.

    This formulation draws an optimization architecture
    where the coupling variables of strongly coupled disciplines is made consistent
    by adding equality constraints on the coupling variables at top level,
    the optimization problem
    with respect to the local, global design variables and coupling variables
    is made at the top level.

    The disciplinary analysis is made at a each optimization iteration
    while the multidisciplinary analysis is made at the optimum.
    """

    def __init__(
        self,
        disciplines,  # type: Sequence[MDODiscipline]
        objective_name,  # type: str
        design_space,  # type: DesignSpace
        maximize_objective=False,  # type: bool
        normalize_constraints=True,  # type: bool
        parallel_exec=False,  # type: bool
        use_threading=True,  # type: bool
        start_at_equilibrium=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            normalize_constraints: If True,
                the outputs of the coupling consistency constraints are scaled.
            parallel_exec: If True,
                all constraints and objectives are computed in parallel.
                At every iteration,
                all disciplines are executed in parallel.
                Otherwise,
                a separate constraint is created for each discipline with couplings.
            use_threading: If True and parallel_exec=True,
                the disciplines are run in parallel using multi-threading.
                If False and parallel_exec=True, multi-processing is used.
            start_at_equilibrium: If True,
                an MDA is used to initialize the coupling variables.
        """
        super(IDF, self).__init__(
            disciplines,
            objective_name,
            design_space,
            maximize_objective=maximize_objective,
        )
        if parallel_exec:
            self._parallel_exec = MDOParallelChain(
                self.disciplines, use_threading=use_threading
            )
        else:
            self._parallel_exec = None

        self.coupling_structure = MDOCouplingStructure(disciplines)
        self.all_couplings = self.coupling_structure.get_all_couplings()
        self._update_design_space()
        self.normalize_constraints = normalize_constraints
        self._build_constraints()
        self._build_objective_from_disc(objective_name)

        if start_at_equilibrium:
            self._compute_equilibrium()

    def _compute_equilibrium(self):  # type: (...) -> None
        """Run an MDA to compute the initial target couplings at equilibrium.

        The values at equilibrium are set in the initial design space.
        """
        current_x = self.design_space.get_current_x_dict()
        # run MDA to initialize target coupling variables
        mda = MDAChain(self.disciplines)
        res = mda.execute(current_x)

        for name in self.all_couplings:
            value = res[name]
            LOGGER.info(
                "IDF: changing the initial value of %s " "from %s to %s (equilibrium)",
                name,
                str(current_x[name]),
                str(value),
            )
            self.design_space.set_current_variable(name, value)

    def _update_design_space(self):
        """Update the design space with the required variables."""
        strong_couplings = set(self.all_couplings)
        variables_names = set(self.opt_problem.design_space.variables_names)
        if not strong_couplings.issubset(variables_names):
            missing = strong_couplings - variables_names
            raise ValueError(
                "IDF formulation needs coupling variables as design variables, "
                "missing variables: %s" % missing
            )
        self._set_defaultinputs_from_ds()

    def get_top_level_disc(self):  # type: (...) -> List[MDODiscipline]
        # All functions and constraints are built from the top level disc
        # If we are in parallel mode: return the parallel execution
        if self._parallel_exec is not None:
            return [self._parallel_exec]
        # Otherwise the disciplines are top level
        return self.disciplines

    def _get_normalization_factor(
        self,
        output_couplings,  # type:Iterable[str]
    ):  # type: (...) -> ndarray
        """Compute [abs(ub-lb)] for all output couplings.

        Args:
            output_couplings: The names of the variables for normalization.

        Returns:
            The concatenation of the normalization factors for all output couplings.
        """
        norm_fact = []
        for output in output_couplings:
            u_b = self.design_space.get_upper_bound(output)
            l_b = self.design_space.get_lower_bound(output)
            norm_fact.append(np_abs(u_b - l_b))
        return concatenate(norm_fact)

    def _generate_consistency_cstr(
        self,
        output_couplings,  # type: Sequence[str]
    ):  # type: (...) -> MDOFunction
        """Generate the consistency constraints for a discipline.

        Args:
            output_couplings: The names of the output couplings.

        Returns:
            A function computing the consistency constraints.
        """
        coupl_func = self._get_function_from(output_couplings)
        dv_names_of_disc = coupl_func.args

        if self.normalize_constraints:
            norm_fact = self._get_normalization_factor(output_couplings)
        else:
            norm_fact = 1.0

        def coupl_min_x(
            x_vec,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Function to compute the consistency constraints.

            Args:
                x_vect: The design variable vector.

            Returns:
                The value of the consistency constraints.
                Equal to zero if the disciplines are at equilibrium.
            """
            x_sw = self.mask_x_swap_order(output_couplings, x_vec)
            coupl = coupl_func(x_vec)
            if self.normalize_constraints:
                return (coupl - x_sw) / norm_fact
            return coupl - x_sw

        def coupl_min_x_jac(
            x_vec,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Function to compute the gradient of the consistency constraints.

            Args:
                x_vect: The design variable vector.

            Returns:
                The value of the gradient of the consistency constraints.
            """
            coupl_jac = coupl_func.jac(x_vec)  # pylint: disable=E1102

            if len(coupl_jac.shape) > 1:
                # IN this case it is harder since a block diagonal
                # matrix with -Id should be placed for each output
                # coupling, at the right place
                n_outs = coupl_jac.shape[0]
                x_jac_2d = zeros((n_outs, len(x_vec)), dtype=x_vec.dtype)
                x_names = self.get_optim_variables_names()
                o_min = 0
                o_max = 0
                for out in output_couplings:
                    # self._reference_input_data[out].size
                    o_len = self._get_dv_length(out)
                    i_min = 0
                    i_max = 0
                    o_max += o_len
                    for x_i in x_names:
                        # self._reference_input_data[x_i].size
                        x_len = self._get_dv_length(x_i)
                        i_max += x_len
                        if x_i == out:
                            x_jac_2d[o_min:o_max, i_min:i_max] = eye(x_len)
                        i_min = i_max
                    o_min = o_max
                x_jac = x_jac_2d
            else:
                # This is surprising but there is a duality between the masking
                # operation in the function inputs and the unmasking of its
                # outputs
                x_jac = self.unmask_x_swap_order(output_couplings, ones_like(x_vec))
            if self.normalize_constraints:
                return (coupl_jac - x_jac) / norm_fact[:, None]
            return coupl_jac - x_jac

        expr = ""
        for out_c in output_couplings:
            expr += out_c + "(" + ", ".join(dv_names_of_disc) + ") - "
            expr += str(out_c) + "" + "\n"

        name = coupl_func.name
        return MDOFunction(
            coupl_min_x,
            name,
            args=dv_names_of_disc,
            expr=expr,
            jac=coupl_min_x_jac,
            outvars=coupl_func.outvars,
            f_type=MDOFunction.TYPE_EQ,
        )

    def _build_constraints(self):  # type: (...) -> None
        """Build the constraints.

        In IDF formulation,
        the consistency constraints are "y - y_copy = 0".
        """
        # Building constraints per generator couplings
        for discipline in self.disciplines:
            couplings = self.coupling_structure.output_couplings(
                discipline, strong=False
            )
            if couplings:
                cstr = self._generate_consistency_cstr(couplings)
                self.opt_problem.add_eq_constraint(cstr)

    def get_expected_workflow(
        self,
    ):  # type: (...) -> List[ExecutionSequence,Tuple[ExecutionSequence]]
        return ExecutionSequenceFactory.parallel(self.disciplines)

    def get_expected_dataflow(
        self,
    ):  # type: (...) -> List[Tuple[MDODiscipline,MDODiscipline,List[str]]]
        return []



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Damien Guenot
#       :author: Francois Gallard, Charlie Vanaret, Benoit Pauwels
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Optimization problem.

The :class:`.OptimizationProblem` class operates on a :class:`.DesignSpace` defining:

- an initial guess :math:`x_0` for the design variables,
- the bounds :math:`l_b \\leq x \\leq u_b` of the design variables.

A (possible vector) objective function with a :class:`.MDOFunction` type
is set using the :code:`objective` attribute.
If the optimization problem looks for the maximum of this objective function,
the :meth:`.OptimizationProblem.change_objective_sign`
changes the objective function sign
because the optimization drivers seek to minimize this objective function.

Equality and inequality constraints are also :class:`.MDOFunction` instances
provided to the :class:`.OptimizationProblem`
by means of its :meth:`.OptimizationProblem.add_constraint` method.

The :class:`.OptimizationProblem` allows to evaluate the different functions
for a given design parameters vector
(see :meth:`.OptimizationProblem.evaluate_functions`).
Note that this evaluation step relies on an automated scaling of function wrt the bounds
so that optimizers and DOE algorithms work
with inputs scaled between 0 and 1 for all the variables.

The :class:`.OptimizationProblem`  has also a :class:`.Database`
that stores the calls to all the functions
so that no function is called twice with the same inputs.
Concerning the derivatives computation,
the :class:`.OptimizationProblem` automates
the generation of the finite differences or complex step wrappers on functions,
when the analytical gradient is not available.

Lastly,
various getters and setters are available,
as well as methods to export the :class:`.Database`
to a HDF file or to a :class:`.Dataset` for future post-processing.
"""
from __future__ import division, unicode_literals

import logging
from functools import reduce
from numbers import Number
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import h5py
import numpy
from numpy import abs as np_abs
from numpy import all as np_all
from numpy import any as np_any
from numpy import argmin, array, array_equal, concatenate, inf, insert
from numpy import isnan as np_isnan
from numpy import issubdtype, multiply, nan, ndarray
from numpy import number as np_number
from numpy import where
from numpy.linalg import norm
from six import string_types

from gemseo.algos.aggregation.aggregation_func import (
    aggregate_iks,
    aggregate_ks,
    aggregate_max,
    aggregate_sum_square,
)
from gemseo.algos.database import Database
from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_result import OptimizationResult
from gemseo.algos.stop_criteria import DesvarIsNan, FunctionIsNan
from gemseo.core.dataset import Dataset
from gemseo.core.function import MDOFunction, MDOLinearFunction, MDOQuadraticFunction
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.derivatives_approx import ComplexStep, FirstOrderFD
from gemseo.utils.hdf5 import get_hdf5_group
from gemseo.utils.py23_compat import PY3, string_array
from gemseo.utils.string_tools import MultiLineString, pretty_repr

LOGGER = logging.getLogger(__name__)

NAME_TO_METHOD = {
    "sum": aggregate_sum_square,
    "max": aggregate_max,
    "KS": aggregate_ks,
    "IKS": aggregate_iks,
}

BestInfeasiblePointType = Tuple[
    Optional[ndarray], Optional[ndarray], bool, Dict[str, ndarray]
]
OptimumType = Tuple[ndarray, ndarray, bool, Dict[str, ndarray], Dict[str, ndarray]]
OptimumSolutionType = Tuple[
    Optional[Sequence[ndarray]], ndarray, Dict[str, ndarray], Dict[str, ndarray]
]


class OptimizationProblem(object):
    """An optimization problem.

    Create an optimization problem from:

    - a :class:`.DesignSpace` specifying the design variables
      in terms of names, lower bounds, upper bounds and initial guesses,
    - the objective function as a :class:`.MDOFunction`,
      which can be a vector,

    execute it from an algorithm provided by a :class:`.DriverLib`,
    and store some execution data in a :class:`.Database`.

    In particular,
    this :class:`.Database` stores the calls to all the functions
    so that no function is called twice with the same inputs.

    An :class:`.OptimizationProblem` also has an automated scaling of function
    with respect to the bounds of the design variables
    so that the driving algorithms work with inputs scaled between 0 and 1.

    Lastly, :class:`.OptimizationProblem` automates the generation
    of finite differences or complex step wrappers on functions,
    when analytical gradient is not available.

    Attributes:
        nonproc_objective (MDOFunction): The non-processed objective function.
        constraints (List(MDOFunction)): The constraints.
        nonproc_constraints (List(MDOFunction)): The non-processed constraints.
        observables (List(MDOFunction)): The observables.
        new_iter_observables (List(MDOFunction)): The observables to be called
            at each new iterate.
        nonproc_observables (List(MDOFunction)): The non-processed observables.
        nonproc_new_iter_observables (List(MDOFunction)): The non-processed observables
            to be called at each new iterate.
        minimize_objective (bool): If True, maximize the objective.
        fd_step (float): The finite differences step.
        differentiation_method (str): The type differentiation method.
        pb_type (str): The type of optimization problem.
        ineq_tolerance (float): The tolerance for the inequality constraints.
        eq_tolerance (float): The tolerance for the equality constraints.
        database (Database): The database to store the optimization problem data.
        solution: The solution of the optimization problem.
        design_space (DesignSpace): The design space on which the optimization problem
            is solved.
        stop_if_nan (bool): If True, the optimization stops when a function returns NaN.
        preprocess_options (Dict): The options to pre-process the functions.
    """

    LINEAR_PB = "linear"
    NON_LINEAR_PB = "non-linear"
    AVAILABLE_PB_TYPES = [LINEAR_PB, NON_LINEAR_PB]

    USER_GRAD = "user"
    COMPLEX_STEP = "complex_step"
    FINITE_DIFFERENCES = "finite_differences"
    NO_DERIVATIVES = "no_derivatives"
    DIFFERENTIATION_METHODS = [
        USER_GRAD,
        COMPLEX_STEP,
        FINITE_DIFFERENCES,
        NO_DERIVATIVES,
    ]
    DESIGN_VAR_NAMES = "x_names"
    DESIGN_VAR_SIZE = "x_size"
    DESIGN_SPACE_ATTRS = ["u_bounds", "l_bounds", "x_0", DESIGN_VAR_NAMES, "dimension"]
    FUNCTIONS_ATTRS = ["objective", "constraints"]
    OPTIM_DESCRIPTION = [
        "minimize_objective",
        "fd_step",
        "differentiation_method",
        "pb_type",
        "ineq_tolerance",
        "eq_tolerance",
    ]

    OPT_DESCR_GROUP = "opt_description"
    DESIGN_SPACE_GROUP = "design_space"
    OBJECTIVE_GROUP = "objective"
    SOLUTION_GROUP = "solution"
    CONSTRAINTS_GROUP = "constraints"

    HDF5_FORMAT = "hdf5"
    GGOBI_FORMAT = "ggobi"

    def __init__(
        self,
        design_space,  # type: DesignSpace
        pb_type=NON_LINEAR_PB,  # type: str
        input_database=None,  # type: Optional[Union[str,Database]]
        differentiation_method=USER_GRAD,  # type: str
        fd_step=1e-7,  # type: float
    ):  # type: (...) -> None
        # noqa: D205, D212, D415
        """
        Args:
            design_space: The design space on which the functions are evaluated.
            pb_type: The type of the optimization problem
                among :attr:`.OptimizationProblem.AVAILABLE_PB_TYPES`.
            input_database: A database to initialize that of the optimization problem.
                If None, the optimization problem starts from an empty database.
            differentiation_method: The default differentiation method to be applied
                to the functions of the optimization problem.
            fd_step: The step to be used by the step-based differentiation methods.
        """
        self._objective = None
        self.nonproc_objective = None
        self.constraints = []
        self.nonproc_constraints = []
        self.observables = []
        self.new_iter_observables = []
        self.nonproc_observables = []
        self.nonproc_new_iter_observables = []
        self.minimize_objective = True
        self.fd_step = fd_step
        self.differentiation_method = differentiation_method
        self.pb_type = pb_type
        self.ineq_tolerance = 1e-4
        self.eq_tolerance = 1e-2
        self.__functions_are_preprocessed = False
        if isinstance(input_database, Database):
            self.database = input_database
        else:
            self.database = Database(input_hdf_file=input_database)
        self.solution = None
        self.design_space = design_space
        self.__x0 = None
        self.stop_if_nan = True
        self.__store_listeners = []
        self.__newiter_listeners = []
        self.preprocess_options = {}

    @property
    def objective(self):  # type: (...) -> MDOFunction
        """The objective function."""
        return self._objective

    @objective.setter
    def objective(
        self,
        func,  # type: MDOFunction
    ):  # type: (...) -> None
        self._objective = func

    @staticmethod
    def repr_constraint(
        func,  # type: MDOFunction
        ctype,  # type: str
        value=None,  # type: Optional[float]
        positive=False,  # type: bool
    ):  # type: (...) -> str
        """Express a constraint as a string expression.

        Args:
            func: The constraint function.
            ctype: The type of the constraint.
                Either equality or inequality.
            value: The value for which the constraint is active.
                If None, this value is 0.
            positive: If True, then the inequality constraint is positive.

        Returns:
            A string representation of the constraint.
        """
        if value is None:
            value = 0.0
        str_repr = func.name
        if func.has_args():
            arguments = ", ".join(func.args)
            str_repr += "({})".format(arguments)

        if ctype == func.TYPE_EQ:
            sign = " == "
        elif positive:
            sign = " >= "
        else:
            sign = " <= "

        if func.has_expr():
            str_repr += ": "
            expr = func.expr
            n_char = len(str_repr)
            # Remove empty lines with filter
            expr_spl = [_f for _f in expr.split("\n") if _f]
            str_repr = str_repr + expr_spl[0] + sign + str(value)
            if isinstance(func, (MDOLinearFunction, MDOQuadraticFunction)):
                for repre in expr_spl[1:]:
                    str_repr += "\n" + " " * n_char + repre
            else:
                for repre in expr_spl[1:]:
                    str_repr += "\n" + " " * n_char + repre + sign + str(value)
        else:
            str_repr += sign + str(value)
        return str_repr

    def add_constraint(
        self,
        cstr_func,  # type: MDOFunction
        value=None,  # type: Optional[value]
        cstr_type=None,  # type: Optional[str]
        positive=False,  # type: bool
    ):  # type: (...) -> None
        """Add a constraint (equality and inequality) to the optimization problem.

        Args:
            cstr_func: The constraint.
            value: The value for which the constraint is active.
                If None, this value is 0.
            cstr_type: The type of the constraint.
                Either equality or inequality.
            positive: If True, then the inequality constraint is positive.
        """
        self.check_format(cstr_func)
        if self.pb_type == OptimizationProblem.LINEAR_PB and not isinstance(
            cstr_func, MDOLinearFunction
        ):
            raise TypeError("Constraint must be an MDOLinearFunction")
        ctype = cstr_type or cstr_func.f_type
        cstr_repr = self.repr_constraint(cstr_func, ctype, value, positive)
        if value is not None:
            cstr_func = cstr_func.offset(-value)
        if positive:
            cstr_func = -cstr_func  # Operator overloading in MDOFunction
        if cstr_type is not None:
            cstr_func.f_type = cstr_type
        else:
            if not cstr_func.is_constraint():
                msg = "Constraint type must be provided, either in the "
                msg += "function attributes or to the add_constraint method"
                raise ValueError(msg)
        cstr_func.special_repr = cstr_repr
        self.constraints.append(cstr_func)

    def add_eq_constraint(
        self,
        cstr_func,  # type: MDOFunction
        value=None,  # type: Optional[float]
    ):  # type: (...) -> None
        """Add an equality constraint to the optimization problem.

        Args:
            cstr_func: The constraint.
            value: The value for which the constraint is active.
                If None, this value is 0.
        """
        self.add_constraint(cstr_func, value, cstr_type=MDOFunction.TYPE_EQ)

    def add_ineq_constraint(
        self,
        cstr_func,  # type: MDOFunction
        value=None,  # type: Optional[value]
        positive=False,  # type: bool
    ):  # type: (...) -> None
        """Add an inequality constraint to the optimization problem.

        Args:
            cstr_func: The constraint.
            value: The value for which the constraint is active.
                If None, this value is 0.
            positive: If True, then the inequality constraint is positive.
        """
        self.add_constraint(
            cstr_func, value, cstr_type=MDOFunction.TYPE_INEQ, positive=positive
        )

    def aggregate_constraint(self, constr_id, method="max", groups=None, **options):
        """Aggregates a constraint to generate a reduced dimension constraint.

        :param constr_id: index of the constraint in self.constraints
        :type constr_id: int
        :param method: aggregation method, among ('max','KS', 'IKS')
        :type method: str or callable, that takes a function and returns a function
        :param groups: if None, a single output constraint is produced
            otherwise, one output per group is produced.
        :type groups: tuple of ndarray
        """
        if constr_id >= len(self.constraints):
            raise ValueError("constr_id must be lower than the number of constraints.")

        constr = self.constraints[constr_id]

        if callable(method):
            method_imp = method
        else:
            method_imp = NAME_TO_METHOD.get(method)
            if method_imp is None:
                raise ValueError("Unknown method {}".format(method))

        del self.constraints[constr_id]

        if groups is None:
            cstr = method_imp(constr, **options)
            self.constraints.insert(constr_id, cstr)
        else:
            cstrs = [method_imp(constr, indx, **options) for indx in groups]
            icstr = self.constraints[:constr_id]
            ecstr = self.constraints[constr_id:]
            self.constraints = icstr + cstrs + ecstr

    def add_observable(
        self,
        obs_func,  # type: MDOFunction
        new_iter=True,  # type: bool
    ):  # type: (...) -> None
        """Add a function to be observed.

        Args:
            obs_func: An observable to be observed.
            new_iter: If True, then the observable will be called at each new iterate.
        """
        self.check_format(obs_func)
        obs_func.f_type = MDOFunction.TYPE_OBS
        self.observables.append(obs_func)
        if new_iter:
            self.new_iter_observables.append(obs_func)

    def get_eq_constraints(self):  # type: (...) -> List[MDOFunction]
        """Retrieve all the equality constraints.

        Returns:
            The equality constraints.
        """

        def is_equality_constraint(
            func,  # type: MDOFunction
        ):  # type: (...) -> bool
            """Check if a function is an equality constraint.

            Args:
                A function.

            Returns:
                True if the function is an equality constraint.
            """
            return func.f_type == MDOFunction.TYPE_EQ

        return list(filter(is_equality_constraint, self.constraints))

    def get_ineq_constraints(self):  # type: (...) -> List[MDOFunction]
        """Retrieve all the inequality constraints.

        Returns:
            The inequality constraints.
        """

        def is_inequality_constraint(
            func,  # type: MDOFunction
        ):  # type: (...) -> bool
            """Check if a function is an inequality constraint.

            Args:
                A function.

            Returns:
                True if the function is an inequality constraint.
            """
            return func.f_type == MDOFunction.TYPE_INEQ

        return list(filter(is_inequality_constraint, self.constraints))

    def get_observable(
        self,
        name,  # type: str
    ):  # type: (...) -> MDOFunction
        """Retrieve an observable from its name.

        Args:
            name: The name of the observable.

        Returns:
            The observable.

        Raises:
            ValueError: If the observable cannot be found.
        """
        try:
            observable = next(obs for obs in self.observables if obs.name == name)
        except StopIteration:
            raise ValueError("Observable {} cannot be found.".format(name))

        return observable

    def get_ineq_constraints_number(self):  # type: (...) -> int
        """Retrieve the number of inequality constraints.

        Returns:
            The number of inequality constraints.
        """
        return len(self.get_ineq_constraints())

    def get_eq_constraints_number(self):  # type: (...) -> int
        """Retrieve the number of equality constraints.

        Returns:
            The number of equality constraints.
        """
        return len(self.get_eq_constraints())

    def get_constraints_number(self):  # type: (...) -> int
        """Retrieve the number of constraints.

        Returns:
            The number of constraints.
        """
        return len(self.constraints)

    def get_constraints_names(self):  # type: (...) -> List[str]
        """Retrieve the names of the constraints.

        Returns:
            The names of the constraints.
        """
        names = [constraint.name for constraint in self.constraints]
        return names

    def get_nonproc_constraints(self):  # type: (...) -> List[MDOFunction]
        """Retrieve the non-processed constraints.

        Returns:
            The non-processed constraints.
        """
        return self.nonproc_constraints

    def get_design_variable_names(self):  # type: (...) -> List[str]
        """Retrieve the names of the design variables.

        Returns:
            The names of the design variables.
        """
        return self.design_space.variables_names

    def get_all_functions(self):  # type: (...) -> List[MDOFunction]
        """Retrieve all the functions of the optimization problem.

        These functions are the constraints, the objective function and the observables.

        Returns:
            All the functions of the optimization problem.
        """
        return [self.objective] + self.constraints + self.observables

    def get_all_functions_names(self):  # type: (...) -> List[str]
        """Retrieve the names of all the function of the optimization problem.

        These functions are the constraints, the objective function and the observables.

        Returns:
            The names of all the functions of the optimization problem.
        """
        return [func.name for func in self.get_all_functions()]

    def get_objective_name(self):  # type: (...) -> str
        """Retrieve the name of the objective function.

        Returns:
            The name of the objective function.
        """
        return self.objective.name

    def get_nonproc_objective(self):  # type: (...) -> MDOFunction
        """Retrieve the non-processed objective function."""
        return self.nonproc_objective

    def has_nonlinear_constraints(self):  # type: (...) -> bool
        """Check if the problem has non-linear constraints.

        Returns:
            True if the problem has equality or inequality constraints.
        """
        return len(self.constraints) > 0

    def __notify_store_listeners(self):  # type: (...) -> None
        """Notify the listeners that a new store has been made in the database."""
        for func in self.__store_listeners:
            func()

    def __notify_newiter_listeners(
        self,
        xvect=None,  # type: Optional[ndarray]
    ):  # type: (...) -> None
        """Notify the listeners that a new iteration is ongoing.

        Args:
            xvect: The values of the design variables.
        """
        for func in self.__newiter_listeners:
            func()

        if xvect is not None:
            for obs in self.new_iter_observables:
                obs(xvect)

    def has_constraints(self):
        """Check if the problem has equality or inequality constraints.

        Returns:
            True if the problem has equality or inequality constraints.
        """
        return self.has_eq_constraints() or self.has_ineq_constraints()

    def has_eq_constraints(self):  # type: (...) -> bool
        """Check if the problem has equality constraints.

        Returns:
            True if the problem has equality constraints.
        """
        return len(self.get_eq_constraints()) > 0

    def has_ineq_constraints(self):  # type: (...) -> bool
        """Check if the problem has inequality constraints.

        Returns:
            True if the problem has inequality constraints.
        """
        return len(self.get_ineq_constraints()) > 0

    def get_x0_normalized(self):  # type: (...) -> ndarray
        """Return the current values of the design variables after normalization.

        Returns:
            The current values of the design variables
            normalized between 0 and 1 from their lower and upper bounds.
        """
        dspace = self.design_space
        return dspace.normalize_vect(dspace.get_current_x())

    def get_dimension(self):  # type: (...) -> int
        """Retrieve the total number of design variables.

        Returns:
            The dimension of the design space.
        """
        return self.design_space.dimension

    @property
    def dimension(self):  # type: (...) -> int
        """The dimension of the design space."""
        return self.design_space.dimension

    @staticmethod
    def check_format(input_function):  # type: (...) -> None
        """Check that a function is an instance of :class:`.MDOFunction`.

        Args:
            input_function: The function to be tested.

        Raises:
            TypeError: If the function is not a :class:`.MDOFunction`.
        """
        if not isinstance(input_function, MDOFunction):
            raise TypeError(
                "Optimization problem functions must be instances of MDOFunction"
            )

    def get_eq_cstr_total_dim(self):  # type: (...) -> int
        """Retrieve the total dimension of the equality constraints.

        This dimension is the sum
        of all the outputs dimensions
        of all the equality constraints.

        Returns:
            The total dimension of the equality constraints.
        """
        return self.__count_cstr_total_dim(MDOFunction.TYPE_EQ)

    def get_ineq_cstr_total_dim(self):  # type: (...) -> int
        """Retrieve the total dimension of the inequality constraints.

        This dimension is the sum
        of all the outputs dimensions
        of all the inequality constraints.

        Returns:
            The total dimension of the inequality constraints.
        """
        return self.__count_cstr_total_dim(MDOFunction.TYPE_INEQ)

    def __count_cstr_total_dim(
        self,
        cstr_type,  # type: str
    ):  # type: (...) -> int
        """Retrieve the total dimension of the constraints.

        This dimension is the sum
        of all the outputs dimensions
        of all the constraints.
        of equality or inequality constraints dimensions
        that is the sum of all outputs dimensions of all constraints.

        Returns:
            The total dimension of the constraints.
        """
        n_cstr = 0
        for constraint in self.constraints:
            if not constraint.has_dim():
                raise ValueError(
                    "Constraint dimension not available yet, "
                    "please call function {} once".format(constraint)
                )
            if constraint.f_type == cstr_type:
                n_cstr += constraint.dim
        return n_cstr

    def get_active_ineq_constraints(
        self,
        x_vect,  # type: ndarray
        tol=1e-6,  # type: float
    ):  # type: (...) -> Dict[str,ndarray]
        """For each constraint, indicate if its different components are active.

        Args:
            x_vect: The vector of design variables.
            tol: The tolerance for deciding whether a constraint is active.

        Returns:
            For each constraint,
            a boolean indicator of activation of its different components.
        """
        self.design_space.check_membership(x_vect)
        normalize = self.preprocess_options.get("normalize", False)
        if normalize:
            x_vect = self.design_space.normalize_vect(x_vect)

        act_funcs = {}
        for func in self.get_ineq_constraints():
            val = np_abs(func(x_vect))
            act_funcs[func] = where(val <= tol, True, False)

        return act_funcs

    def __wrap_in_database(
        self,
        orig_func,  # type: MDOFunction
        normalized=True,  # type: bool
        is_observable=False,  # type: bool
    ):  # type: (...) -> MDOFunction
        """Wrap a function, test its presence in the database and store its evaluation.

        Args:
            orig_func: The function to be wrapped in the database.
            normalized: If True, the input of the function are assumed normalized.
            is_observable: If True, new_iter_listeners are not called
                when function is called (avoid recursive call)

        Returns:
            The wrapped function.
        """
        fname = orig_func.name
        normalize_vect = self.design_space.normalize_vect
        unnormalize_vect = self.design_space.unnormalize_vect

        def unnormalize_gradient(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Unnormalize the gradient value.

            Args:
                x_vect: The normalized value of the gradient.

            Returns:
                The unnormalized value of the gradient.
            """
            return normalize_vect(x_vect, minus_lb=False)

        def normalize_gradient(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Normalize the gradient value.

            Args:
                x_vect: The unnormalized value of the gradient.

            Returns:
                The normalized value of the gradient.
            """
            return unnormalize_vect(x_vect, minus_lb=False, no_check=True)

        def wrapped_function(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Wrapped provided function in order to give to optimizer.

            Args:
                x_vect: The value of the design variables.

            Returns:
                The evaluation of the function for this value of design variables.
            """
            if np_any(np_isnan(x_vect)):
                raise DesvarIsNan(
                    "Design Variables contain a NaN value ! {}".format(x_vect)
                )
            if normalized:
                xn_vect = x_vect
                xu_vect = unnormalize_vect(xn_vect)
            else:
                xu_vect = x_vect
                xn_vect = normalize_vect(xu_vect)
            # try to retrieve the evaluation
            value = None
            if not self.database.get(xu_vect, False) and not is_observable:
                if normalized:
                    self.__notify_newiter_listeners(xn_vect)
                else:
                    self.__notify_newiter_listeners(xu_vect)
            else:
                value = self.database.get_f_of_x(fname, xu_vect)
            if value is None:
                # if not evaluated yet, evaluate
                if normalized:
                    value = orig_func(xn_vect)
                else:
                    value = orig_func(xu_vect)
                if self.stop_if_nan and np_any(np_isnan(value)):
                    raise FunctionIsNan(
                        "Function {} is NaN for x={}".format(fname, xu_vect)
                    )
                values_dict = {fname: value}
                # store (x, f(x)) in database
                self.database.store(xu_vect, values_dict)
                self.__notify_store_listeners()

            return value

        db_func = MDOFunction(
            wrapped_function,
            name=fname,
            f_type=orig_func.f_type,
            expr=orig_func.expr,
            args=orig_func.args,
            dim=orig_func.dim,
            outvars=orig_func.outvars,
        )

        if orig_func.has_jac():

            def dwrapped_function(
                x_vect,  # type: ndarray
            ):  # type: (...) -> ndarray
                """Wrapped provided gradient in order to give to optimizer.

                Args:
                    x_vect: The value of the design variables.

                Returns:
                    The evaluation of the gradient for this value of design variables.
                """
                if np_any(np_isnan(x_vect)):
                    raise FunctionIsNan(
                        "Design Variables contain a NaN value! {}".format(x_vect)
                    )
                if normalized:
                    xn_vect = x_vect
                    xu_vect = unnormalize_vect(xn_vect)
                else:
                    xu_vect = x_vect
                    xn_vect = normalize_vect(xu_vect)
                # try to retrieve the evaluation
                jac_u = None
                if not self.database.get(xu_vect, False):
                    self.__notify_newiter_listeners()
                else:
                    jac_u = self.database.get_f_of_x(
                        Database.get_gradient_name(fname), xu_vect
                    )

                if jac_u is not None:
                    jac_n = normalize_gradient(jac_u)
                else:
                    # if not evaluated yet, evaluate
                    if normalized:
                        jac_n = orig_func.jac(xn_vect).real
                        jac_u = unnormalize_gradient(jac_n)
                    else:
                        jac_u = orig_func.jac(xu_vect).real
                        jac_n = normalize_gradient(jac_u)
                    if np_any(np_isnan(jac_n)) and self.stop_if_nan:
                        raise FunctionIsNan(
                            "Function {}'s Jacobian is NaN "
                            "for x={}".format(fname, xu_vect)
                        )
                    values_dict = {Database.get_gradient_name(fname): jac_u}
                    # store (x, j(x)) in database
                    self.database.store(xu_vect, values_dict)
                    self.__notify_store_listeners()
                if normalized:
                    jac = jac_n
                else:
                    jac = jac_u
                return jac

            db_func.jac = dwrapped_function

        return db_func

    def add_callback(
        self,
        callback_func,  # type: Callable
        each_new_iter=True,  # type: bool
        each_store=False,  # type: bool
    ):  # type: (...) -> None
        """Add a callback function after each store operation or new iteration.

        Args:
            callback_func: A function to be called after some event.
            each_new_iter: If True, then callback at every iteration.
            each_store: If True,
                then callback at every call to :class:`.Database.store`.
        """
        if each_store:
            self.add_store_listener(callback_func)
        if each_new_iter:
            self.add_new_iter_listener(callback_func)

    def add_store_listener(
        self,
        listener_func,  # type: Callable
    ):  # type: (...) -> None
        """Add a listener to be called when an item is stored to the database.

        Args:
            listener_func: The function to be called.

        Raises:
            TypeError: If the argument is not a callable
        """
        if not callable(listener_func):
            raise TypeError("Listener function is not callable")
        self.__store_listeners.append(listener_func)

    def add_new_iter_listener(
        self,
        listener_func,  # type: Callable
    ):  # type: (...) -> None
        """Add a listener to be called when a new iteration is stored to the database.

        Args:
            listener_func: The function to be called.

        Raises:
            TypeError: If the argument is not a callable
        """
        if not callable(listener_func):
            raise TypeError("Listener function is not callable")
        self.__newiter_listeners.append(listener_func)

    def clear_listeners(self):  # type: (...) -> None
        """Clear all the listeners."""
        self.__store_listeners = []
        self.__newiter_listeners = []

    def evaluate_functions(
        self,
        x_vect=None,  # type: ndarray
        eval_jac=False,  # type: bool
        eval_obj=True,  # type:bool
        normalize=True,  # type:bool
        no_db_no_norm=False,  # type:bool
    ):  # type: (...) -> Tuple[Dict[str,Union[float,ndarray]],Dict[str,ndarray]]
        """Compute the objective and the constraints.

        Some optimization libraries require the number of constraints
        as an input parameter which is unknown by the formulation or the scenario.
        Evaluation of initial point allows to get this mandatory information.
        This is also used for design of experiments to evaluate samples.

        Args:
            x_vect: The input vector at which the functions must be evaluated;
                if None, x_0 is used.
            eval_jac: If True, then the Jacobian is evaluated
            eval_obj: If True, then the objective function is evaluated
            normalize: If True, then input vector is considered normalized
            no_db_no_norm: If True, then do not use the pre-processed functions,
                so we have no database, nor normalization.

        Returns:
            The functions values and/or the Jacobian values
            according to the passed arguments.

        Raises:
            ValueError: If both no_db_no_norm and normalize are True.
        """
        if no_db_no_norm and normalize:
            raise ValueError("Can't use no_db_no_norm and normalize options together")
        if normalize:
            if x_vect is None:
                x_vect = self.get_x0_normalized()
            else:
                # Checks proposed x wrt bounds
                x_u_r = self.design_space.unnormalize_vect(x_vect)
                self.design_space.check_membership(x_u_r)
        else:
            if x_vect is None:
                x_vect = self.design_space.get_current_x()
            else:
                # Checks proposed x wrt bounds
                self.design_space.check_membership(x_vect)

        if no_db_no_norm:
            if eval_obj:
                functions = self.nonproc_constraints + [self.nonproc_objective]
            else:
                functions = self.nonproc_constraints
        else:
            if eval_obj:
                functions = self.constraints + [self.objective]
            else:
                functions = self.constraints

        outputs = {}
        for func in functions:
            try:
                outputs[func.name] = func(x_vect)
            except ValueError:
                LOGGER.error("Failed to evaluate function %s", func.name)
                raise

        jacobians = {}
        if eval_jac:
            for func in functions:
                try:
                    jacobians[func.name] = func.jac(x_vect)
                except ValueError:
                    msg = "Failed to evaluate jacobian of {}".format(func.name)
                    LOGGER.error(msg)
                    raise

        return outputs, jacobians

    def preprocess_functions(
        self,
        normalize=True,  # type: bool
        use_database=True,  # type: bool
        round_ints=True,  # type: bool
    ):  # type: (...) -> None
        """Pre-process all the functions and eventually the gradien.

        Required to wrap the objective function and constraints with the database
        and eventually the gradients by complex step or finite differences.

        Args:
            normalize: If True, then the functions are normalized.
            use_database: If True, then the functions are wrapped in the database.
            round_ints: If True, then round the integer variables.
        """
        if round_ints:
            # Keep the rounding option only if there is an integer design variable
            round_ints = any(
                (
                    np_any(var_type == DesignSpace.INTEGER)
                    for var_type in self.design_space.variables_types.values()
                )
            )
        # Avoids multiple wrappings of functions when multiple executions
        # are performed, in bi level scenarios for instance
        if not self.__functions_are_preprocessed:
            self.preprocess_options = {
                "normalize": normalize,
                "use_database": use_database,
                "round_ints": round_ints,
            }
            # Preprocess the constraints
            for icstr, cstr in enumerate(self.constraints):
                self.nonproc_constraints.append(cstr)
                p_cstr = self.__preprocess_func(
                    cstr,
                    normalize=normalize,
                    use_database=use_database,
                    round_ints=round_ints,
                )
                p_cstr.special_repr = cstr.special_repr
                self.constraints[icstr] = p_cstr
            # Preprocess the observables
            for iobs, obs in enumerate(self.observables):
                self.nonproc_observables.append(obs)
                p_obs = self.__preprocess_func(
                    obs,
                    normalize=normalize,
                    use_database=use_database,
                    round_ints=round_ints,
                    is_observable=True,
                )
                p_obs.special_repr = obs.special_repr
                self.observables[iobs] = p_obs
            for iobs, obs in enumerate(self.new_iter_observables):
                self.nonproc_new_iter_observables.append(obs)
                p_obs = self.__preprocess_func(
                    obs,
                    normalize=normalize,
                    use_database=use_database,
                    round_ints=round_ints,
                    is_observable=True,
                )
                p_obs.special_repr = obs.special_repr
                self.new_iter_observables[iobs] = p_obs
            # Preprocess the objective
            self.nonproc_objective = self.objective
            self.objective = self.__preprocess_func(
                self.objective,
                normalize=normalize,
                use_database=use_database,
                round_ints=round_ints,
            )
            self.objective.special_repr = self.objective.special_repr
            self.objective.f_type = MDOFunction.TYPE_OBJ
            self.__functions_are_preprocessed = True
            self.check()

    def __preprocess_func(
        self,
        function,  # type: MDOFunction
        normalize=True,  # type: bool
        use_database=True,  # type: bool
        round_ints=True,  # type: bool
        is_observable=False,  # type: bool
    ):  # type: (...) -> MDOFunction
        """Wrap the function to differentiate it and store its call in the database.

        Only the computed gradients are stored in the database,
        not the eventual finite differences or complex step perturbed evaluations.

        Args:
            function: The scaled and derived function to be pre-processed.
            normalize: If True, then the function is normalized.
            use_database: If True, then the function is wrapped in the database.
            round_ints: If True, then round the integer variables.
            is_observable: If True, new_iter_listeners are not called
                when function is called (avoid recursive call)

        Returns:
            The preprocessed function.
        """
        self.check_format(function)
        # First differentiate it so that the finite differences evaluations
        # are not stored in the database, which would be the case in the other
        # way round
        # Also, store non normalized values in the database for further
        # exploitation
        if isinstance(function, MDOLinearFunction) and not round_ints and normalize:
            function = self.__normalize_linear_function(function)
        else:
            function = self.__normalize_and_round(function, normalize, round_ints)
        self.__add_fd_jac(function)

        # Cast to real value, the results can be a complex number (ComplexStep)
        function.force_real = True
        if use_database:
            function = self.__wrap_in_database(function, normalize, is_observable)
        return function

    def __normalize_and_round(
        self,
        orig_func,  # type: MDOFunction
        normalize,  # type: bool
        round_ints,  # type: bool
    ):  # type: (...) -> MDOFunction
        """Create a function using a scaled input vector.

        Args:
            orig_func: The original function.
            normalize: If True, then the function is normalized.
            round_ints: If True, then round the integer variables.

        Returns:
            A normalized and differentiated function.
        """
        if (not normalize) and (not round_ints):
            return orig_func

        unnormalize_vect = self.design_space.unnormalize_vect

        def normalize_gradient(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Normalize a gradient.

            Args:
                x_vect: The gradient to be normalized.

            Returns:
                The normalized gradient.
            """
            return unnormalize_vect(x_vect, minus_lb=False, no_check=True)

        round_int_vars = self.design_space.round_vect

        def f_wrapped(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Evaluate the original function.

            Args:
                x_vect: An input vector.

            Returns:
                The value of the function at this input vector.
            """
            if normalize:
                x_vect = unnormalize_vect(x_vect)
            if round_ints:
                x_vect = round_int_vars(x_vect)
            return orig_func(x_vect)

        normed_func = MDOFunction(
            f_wrapped,
            name=orig_func.name,
            f_type=orig_func.f_type,
            expr=orig_func.expr,
            args=orig_func.args,
            dim=orig_func.dim,
            outvars=orig_func.outvars,
        )

        def df_wrapped(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Evaluate the gradient of the original function.

            Args:
                x_vect: An input vector.

            Returns:
                The value of the gradient of the original function at this input vector.

            Raises:
                ValueError: If the original function does not provide a Jacobian matrix.
            """
            if not orig_func.has_jac():
                raise ValueError(
                    "Selected user gradient but function {} "
                    "has no Jacobian matrix !".format(orig_func)
                )
            if normalize:
                x_vect = unnormalize_vect(x_vect)
            if round_ints:
                x_vect = round_int_vars(x_vect)
            g_u = orig_func.jac(x_vect)
            if normalize:
                return normalize_gradient(g_u)
            return g_u

        normed_func.jac = df_wrapped

        return normed_func

    def __normalize_linear_function(
        self,
        orig_func,  # type: MDOLinearFunction
    ):  # type: (...) -> MDOLinearFunction
        """Create a linear function using a scaled input vector.

        Args:
            orig_func: The original linear function

        Returns:
            The scaled linear function.

        Raises:
            TypeError: If the original function is not a :class:`.MDOLinearFunction`.
        """
        if not isinstance(orig_func, MDOLinearFunction):
            raise TypeError("Original function must be linear")
        design_space = self.design_space

        # Get normalization factors and shift
        norm_policies = design_space.dict_to_array(design_space.normalize)
        norm_factors = where(
            norm_policies,
            design_space.get_upper_bounds() - design_space.get_lower_bounds(),
            1.0,
        )
        shift = where(norm_policies, design_space.get_lower_bounds(), 0.0)

        # Build the normalized linear function
        coefficients = multiply(orig_func.coefficients, norm_factors)
        value_at_zero = orig_func(shift)
        normalized_func = MDOLinearFunction(
            coefficients,
            orig_func.name,
            orig_func.f_type,
            orig_func.args,
            value_at_zero,
        )

        return normalized_func

    def __add_fd_jac(
        self,
        function,  # type: MDOFunction
    ):  # type: (...) -> None
        """Add a pointer to the approached Jacobian of the function.

        This Jacobian matrix is generated either by COMPLEX_STEP or FINITE_DIFFERENCES.

        Args:
            function: The function to be derivated.
        """
        if self.differentiation_method == self.COMPLEX_STEP:
            c_s = ComplexStep(function.evaluate, self.fd_step)
            function.jac = c_s.f_gradient

        if self.differentiation_method == self.FINITE_DIFFERENCES:
            f_d = FirstOrderFD(function, self.fd_step)
            function.jac = f_d.f_gradient

    def check(self):  # type: (...) -> None
        """Check if the optimization problem is ready for run.

        Raises:
            ValueError: If the objective function is missing.
        """
        if self.objective is None:
            raise ValueError("Missing objective function in OptimizationProblem")
        self.__check_pb_type()
        self.design_space.check()
        self.__check_differentiation_method()
        self.check_format(self.objective)
        self.__check_functions()

    def __check_functions(self):  # type: (...) -> None
        """Check that the constraints are well declared.

        Raises:
            ValueError: If a function declared as a constraint has the wrong type.
        """
        for cstr in self.constraints:
            self.check_format(cstr)
            if not cstr.is_constraint():
                raise ValueError(
                    "Constraint type is not eq or ineq !, got {}"
                    " instead ".format(cstr.f_type)
                )
        self.check_format(self.objective)

    def __check_pb_type(self):
        """Check that the problem type is right.

        Available type are: :attr:`.OptimizationProblem.AVAILABLE_PB_TYPES`.

        Raises:
            ValueError: If a function declared as a constraint has the wrong type.
        """
        if self.pb_type not in self.AVAILABLE_PB_TYPES:
            raise TypeError(
                "Unknown problem type {}, "
                "available problem types are {}".format(
                    self.pb_type, self.AVAILABLE_PB_TYPES
                )
            )

    def __check_differentiation_method(self):
        """Check that the differentiation method is in allowed ones.

        Available ones are: :attr:`.OptimizationProblem.DIFFERENTIATION_METHODS`.

        Raises:
            ValueError: If either
                the differentiation method is unknown,
                the complex step is null or
                the finite differences step is null.
        """
        if self.differentiation_method not in self.DIFFERENTIATION_METHODS:
            raise ValueError(
                "Unknown method {} "
                "is not among the supported ones: {}".format(
                    self.differentiation_method, self.DIFFERENTIATION_METHODS
                )
            )

        if self.differentiation_method == self.COMPLEX_STEP:
            if self.fd_step == 0:
                raise ValueError("ComplexStep step is null!")
            if self.fd_step.imag != 0:
                LOGGER.warning(
                    "Complex step method has an imaginary "
                    "step while required a pure real one."
                    " Auto setting the real part"
                )
                self.fd_step = self.fd_step.imag
        elif self.differentiation_method == self.FINITE_DIFFERENCES:
            if self.fd_step == 0:
                raise ValueError("Finite differences step is null!")
            if self.fd_step.imag != 0:
                LOGGER.warning(
                    "Finite differences method has a complex "
                    "step while required a pure real one."
                    " Auto setting the imaginary part to 0"
                )
                self.fd_step = self.fd_step.real

    def change_objective_sign(self):  # type: (...) -> None
        """Change the objective function sign in order to minimize its opposite.

        The :class:`.OptimizationProblem` expresses any optimization problem as a
        minimization problem. Then, an objective function originally expressed as a
        performance function to maximize must be converted into a cost function to
        minimize, by means of this method.
        """
        self.minimize_objective = not self.minimize_objective
        self.objective = -self.objective

    def _satisfied_constraint(
        self,
        cstr_type,  # type: str
        value,  # type: ndarray
    ):  # type: (...) -> bool
        """Determine if an evaluation satisfies a constraint within a given tolerance.

        Args:
            cstr_type: The type of the constraint.
            value: The value of the constraint.

        Returns:
            Whether a value satisfies a constraint.
        """
        if cstr_type == MDOFunction.TYPE_EQ:
            return np_all(np_abs(value) <= self.eq_tolerance)
        return np_all(value <= self.ineq_tolerance)

    def is_point_feasible(
        self,
        out_val,  # type: Dict[str,ndarray]
        constraints=None,  # type: Optional[Iterable[MDOFunction]]
    ):  # type: (...) -> bool
        """Check if a point is feasible.

        Note:
            If the value of a constraint is absent from this point,
            then this constraint will be considered satisfied.

        Args:
            out_val: The values of the objective function, and eventually constraints.
            constraints: The constraints whose values are to be tested.
                If None, then take all constraints of the problem.

        Returns:
            The feasibility of the point.
        """
        if constraints is None:
            constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        feasible = True
        for constraint in constraints:
            # look for the evaluation of the constraint
            eval_cstr = out_val.get(constraint.name, None)
            # if evaluation exists, check if it is satisfied
            if eval_cstr is None or not self._satisfied_constraint(
                constraint.f_type, eval_cstr
            ):
                feasible = False
                break
        return feasible

    def get_feasible_points(
        self,
    ):  # type: (...) -> Tuple[List[ndarray],List[Dict[str,Union[float,List[int]]]]]
        """Retrieve the feasible points within a given tolerance.

        This tolerance is defined by
        :attr:`.OptimizationProblem.eq_tolerance` for equality constraints and
        :attr:`.OptimizationProblem.ineq_tolerance` for inequality ones.

        Returns:
            The values of the design variables and objective function
            for the feasible points.
        """
        x_history = []
        f_history = []
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()

        for x_vect, out_val in self.database.items():
            feasible = self.is_point_feasible(out_val, constraints=constraints)

            # if all constraints are satisfied, store the vector
            if feasible:
                x_history.append(x_vect.unwrap())
                f_history.append(out_val)
        return x_history, f_history

    def get_violation_criteria(
        self,
        x_vect,  # type: ndarray
    ):  # type: (...) -> Tuple[bool,float]
        """Compute a violation measure associated to an iteration.

        For each constraint,
        when it is violated,
        add the absolute distance to zero,
        in L2 norm.

        If 0, all constraints are satisfied

        Args:
            x_vect: The vector of the design variables values.

        Returns:
            The feasibility of the point and the violation measure.
        """
        f_violation = 0.0
        is_pt_feasible = True
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        out_val = self.database.get(x_vect)
        for constraint in constraints:
            # look for the evaluation of the constraint
            eval_cstr = out_val.get(constraint.name, None)
            # if evaluation exists, check if it is satisfied
            if eval_cstr is None:
                break
            if not self._satisfied_constraint(constraint.f_type, eval_cstr):
                is_pt_feasible = False
                if constraint.f_type == MDOFunction.TYPE_INEQ:
                    if isinstance(eval_cstr, ndarray):
                        viol_inds = where(eval_cstr > self.ineq_tolerance)
                        f_violation += (
                            norm(eval_cstr[viol_inds] - self.ineq_tolerance) ** 2
                        )
                    else:
                        f_violation += (eval_cstr - self.ineq_tolerance) ** 2
                else:
                    f_violation += norm(abs(eval_cstr) - self.eq_tolerance) ** 2
        return is_pt_feasible, f_violation

    def get_best_infeasible_point(
        self,
    ):  # type: (...) -> BestInfeasiblePointType
        """Retrieve the best infeasible point within a given tolerance.

        Returns:
            The best infeasible point expressed as
            the design variables values,
            the objective function value,
            the feasibility of the point
            and the functions values.
        """
        x_history = []
        f_history = []
        is_feasible = []
        viol_criteria = []
        for x_vect, out_val in self.database.items():
            is_pt_feasible, f_violation = self.get_violation_criteria(x_vect)
            is_feasible.append(is_pt_feasible)
            viol_criteria.append(f_violation)
            x_history.append(x_vect.unwrap())
            f_history.append(out_val)

        is_opt_feasible = False
        if viol_criteria:
            best_i = int(argmin(array(viol_criteria)))
            is_opt_feasible = is_feasible[best_i]
        else:
            best_i = 0

        opt_f_dict = {}
        if len(f_history) <= best_i:
            f_opt = None
            x_opt = None
        else:
            f_opt = f_history[best_i].get(self.objective.name)
            x_opt = x_history[best_i]
            opt_f_dict = f_history[best_i]
        return x_opt, f_opt, is_opt_feasible, opt_f_dict

    def __get_optimum_infeas(
        self,
    ):  # type: (...) -> OptimumSolutionType
        """Retrieve the optimum solution.

        Use a feasibility tolerance, when there is no feasible point.

        Returns:
            The optimum solution expressed by
            the design variables values,
            the objective function value,
            the constraints values
            and the constraints gradients values.
        """
        msg = (
            "Optimization found no feasible point ! "
            + " The least infeasible point is selected."
        )
        LOGGER.warning(msg)
        x_opt, f_opt, _, f_history = self.get_best_infeasible_point()
        c_opt = {}
        c_opt_grad = {}
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        for constraint in constraints:
            c_opt[constraint.name] = f_history.get(constraint.name)
            f_key = Database.get_gradient_name(constraint.name)
            c_opt_grad[constraint.name] = f_history.get(f_key)
        return x_opt, f_opt, c_opt, c_opt_grad

    def __get_optimum_feas(
        self,
        feas_x,  # type: Sequence[ndarray]
        feas_f,  # type: Sequence[Dict[str, Union[float, List[int]]]]
    ):  # type: (...) -> OptimumSolutionType
        """Retrieve the optimum solution.

        Use a feasibility tolerances, when there is a feasible point.

        Args:
            feas_x: The values of the design parameters for the feasible evaluations.
            feas_f: The values of the functions for the feasible evaluations.

        Returns:
            The optimum solution expressed by
            the design variables values,
            the objective function value,
            the constraints values
            and the constraints gradients values.
        """
        f_opt, x_opt = inf, None
        c_opt = {}
        c_opt_grad = {}
        obj_name = self.objective.name
        constraints = self.get_ineq_constraints() + self.get_eq_constraints()
        for (i, out_val) in enumerate(feas_f):
            obj_eval = out_val.get(obj_name)
            if obj_eval is None or isinstance(obj_eval, Number) or obj_eval.size == 1:
                tmp_objeval = obj_eval
            else:
                tmp_objeval = norm(obj_eval)
            if tmp_objeval is not None and tmp_objeval < f_opt:
                f_opt = tmp_objeval
                x_opt = feas_x[i]
                for constraint in constraints:
                    c_name = constraint.name
                    c_opt[c_name] = feas_f[i].get(c_name)
                    c_key = Database.get_gradient_name(c_name)
                    c_opt_grad[constraint.name] = feas_f[i].get(c_key)
        return x_opt, f_opt, c_opt, c_opt_grad

    def get_optimum(self):  # type: (...) -> OptimumType
        """Return the optimum solution within a given feasibility tolerances.

        Returns:
            The optimum result,
            defined by:

            - the value of the objective function,
            - the value of the design variables,
            - the indicator of feasibility of the optimal solution,
            - the value of the constraints,
            - the value of the gradients of the constraints.
        """
        if not self.database:
            raise ValueError("Optimization history is empty")
        feas_x, feas_f = self.get_feasible_points()

        if not feas_x:
            is_feas = False
            x_opt, f_opt, c_opt, c_opt_d = self.__get_optimum_infeas()
        else:
            is_feas = True
            x_opt, f_opt, c_opt, c_opt_d = self.__get_optimum_feas(feas_x, feas_f)
        return f_opt, x_opt, is_feas, c_opt, c_opt_d

    def __repr__(self):  # type: (...) -> str
        msg = MultiLineString()
        msg.add("Optimization problem:")
        msg.indent()
        # objective representation
        minimize_str = "Minimize: "
        n_char = len(minimize_str)
        objective_repr = repr(self.objective)
        obj_repr_lines = [line for line in objective_repr.split("\n") if line]
        msg.add(minimize_str + obj_repr_lines[0])
        for line in obj_repr_lines[1:]:
            msg.add(" " * n_char + line)
        # variables representation
        msg.add("With respect to: {}", pretty_repr(self.design_space.variables_names))
        if self.has_constraints():
            msg.add("Subject to constraints:")
            msg.indent()
            for constraints in self.get_ineq_constraints():
                constraints = [cstr for cstr in str(constraints).split("\n") if cstr]
                for constraint in constraints:
                    msg.add(constraint)
            for constraints in self.get_eq_constraints():
                constraints = [cstr for cstr in str(constraints).split("\n") if cstr]
                for constraint in constraints:
                    msg.add(constraint)
        return str(msg)

    @staticmethod
    def __store_h5data(
        group,  # type: Any
        data_array,  # type: ndarray
        dataset_name,  # type: str
        dtype=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """Store an array in a hdf5 file group.

        Args:
            group: The group pointer.
            data_array: The data to be stored.
            dataset_name: The name of the dataset to store the array.
            dtype: Numpy dtype or string. If None, dtype('f') will be used.
        """
        if data_array is None:
            return
        if isinstance(data_array, ndarray):
            data_array = data_array.real
        if isinstance(data_array, string_types):
            data_array = string_array([data_array])
        if isinstance(data_array, list):
            all_str = reduce(
                lambda x, y: x or y,
                (isinstance(data, string_types) for data in data_array),
            )
            if all_str:
                data_array = string_array(data_array)
                dtype = data_array.dtype
        group.create_dataset(dataset_name, data=data_array, dtype=dtype)

    @classmethod
    def __store_attr_h5data(
        cls,
        obj,  # type: Any
        group,  # type: str
    ):  # type: (...) -> None
        """Store an object that has a get_data_dict_repr attribute in the hdf5 dataset.

        Args:
            obj: The object to store
            group: The hdf5 group.
        """
        data_dict = obj.get_data_dict_repr()
        for attr_name, attr in data_dict.items():
            dtype = None
            is_arr_n = isinstance(attr, ndarray) and issubdtype(attr.dtype, np_number)
            if isinstance(attr, string_types):
                attr = attr.encode("ascii", "ignore")
            elif isinstance(attr, bytes):
                attr = attr.decode()
            elif hasattr(attr, "__iter__") and not is_arr_n:
                if PY3:
                    attr = [
                        att.encode("ascii", "ignore")
                        if isinstance(att, string_types)
                        else att
                        for att in attr
                    ]
                dtype = h5py.special_dtype(vlen=str)
            cls.__store_h5data(group, attr, attr_name, dtype)

    def export_hdf(
        self,
        file_path,  # type: str
        append=False,  # type: bool
    ):  # type: (...) -> None
        """Export the optimization problem to an HDF file.

        Args:
            file_path: The file to store the data.
            append: If True, then the data are appended to the file if not empty.
        """
        LOGGER.info("Export optimization problem to file: %s", str(file_path))

        mode = "a" if append else "w"

        with h5py.File(file_path, mode) as h5file:
            no_design_space = DesignSpace.DESIGN_SPACE_GROUP not in h5file

            if not append or self.OPT_DESCR_GROUP not in h5file:
                opt_group = h5file.require_group(self.OPT_DESCR_GROUP)

                for attr_name in self.OPTIM_DESCRIPTION:
                    attr = getattr(self, attr_name)
                    self.__store_h5data(opt_group, attr, attr_name)

                obj_group = h5file.require_group(self.OBJECTIVE_GROUP)
                self.__store_attr_h5data(self.objective, obj_group)

                if self.constraints:
                    constraint_group = h5file.require_group(self.CONSTRAINTS_GROUP)
                    for constraint in self.constraints:
                        c_subgroup = constraint_group.require_group(constraint.name)
                        self.__store_attr_h5data(constraint, c_subgroup)

                if hasattr(self.solution, "get_data_dict_repr"):
                    sol_group = h5file.require_group(self.SOLUTION_GROUP)
                    self.__store_attr_h5data(self.solution, sol_group)

        self.database.export_hdf(file_path, append=True)

        # Design space shall remain the same in append mode
        if not append or no_design_space:
            self.design_space.export_hdf(file_path, append=True)

    @classmethod
    def import_hdf(
        cls,
        file_path,  # type: str
        x_tolerance=0.0,  # type: float
    ):  # type: (...) -> OptimizationProblem
        """Import an optimization history from an HDF file.

        Args:
            file_path: The file containing the optimization history.
            x_tolerance: The tolerance on the design variables when reading the file.

        Returns:
            The read optimization problem.
        """
        LOGGER.info("Import optimization problem from file: %s", str(file_path))

        design_space = DesignSpace(file_path)
        opt_pb = OptimizationProblem(design_space, input_database=file_path)

        with h5py.File(file_path, "r") as h5file:
            group = get_hdf5_group(h5file, opt_pb.OPT_DESCR_GROUP)

            for attr_name, attr in group.items():
                val = attr[()]
                val = val.decode() if isinstance(val, bytes) else val
                setattr(opt_pb, attr_name, val)

            if opt_pb.SOLUTION_GROUP in h5file:
                data_dict = cls.__h5_group_to_dict(h5file, opt_pb.SOLUTION_GROUP)
                attr = OptimizationResult.init_from_dict_repr(**data_dict)
                opt_pb.solution = attr

            data_dict = cls.__h5_group_to_dict(h5file, opt_pb.OBJECTIVE_GROUP)
            attr = MDOFunction.init_from_dict_repr(**data_dict)

            # The generated functions can be called at the x stored in
            # the database
            attr.set_pt_from_database(
                opt_pb.database, design_space, jac=True, x_tolerance=x_tolerance
            )
            opt_pb.objective = attr

            if opt_pb.CONSTRAINTS_GROUP in h5file:
                group = get_hdf5_group(h5file, opt_pb.CONSTRAINTS_GROUP)

                for cstr_name in group.keys():
                    data_dict = cls.__h5_group_to_dict(group, cstr_name)
                    attr = MDOFunction.init_from_dict_repr(**data_dict)
                    opt_pb.constraints.append(attr)

        return opt_pb

    def export_to_dataset(
        self,
        name,  # type: str
        by_group=True,  # type: bool
        categorize=True,  # type: bool
        opt_naming=True,  # type: bool
        export_gradients=False,  # type: bool
    ):  # type: (...) -> Dataset
        """Export the database of the optimization problem to a :class:`.Dataset`.

        The variables can be classified into groups,
        separating the design variables and functions
        (objective function and constraints).
        This classification can use either an optimization naming,
        with :attr:`.Database.DESIGN_GROUP` and :attr:`.Database.FUNCTION_GROUP`
        or an input-output naming,
        with :attr:`.Database.INPUT_GROUP` and :attr:`.Database.OUTPUT_GROUP`

        Args:
            name: A name to be given to the dataset.
            by_group: If True, then store the data by group.
                Otherwise, store them by variables.
            categorize: If True, then distinguish
                between the different groups of variables.
            opt_naming: If True, then use an optimization naming.
            export_gradients: If True, then export also the gradients of the functions
                (objective function, constraints and observables)
                if the latter are available in the database of the optimization problem.

        Returns:
            A dataset built from the database of the optimization problem.
        """
        dataset = Dataset(name, by_group)

        # Set the different groups
        input_group = output_group = gradient_group = dataset.DEFAULT_GROUP
        cache_output_as_input = True
        if categorize:
            if opt_naming:
                input_group = dataset.DESIGN_GROUP
                output_group = dataset.FUNCTION_GROUP
                gradient_group = dataset.GRADIENT_GROUP
            else:
                input_group = dataset.INPUT_GROUP
                output_group = dataset.OUTPUT_GROUP
                gradient_group = dataset.GRADIENT_GROUP
            cache_output_as_input = False

        # Add database inputs
        inputs_names = self.design_space.variables_names
        sizes = self.design_space.variables_sizes
        inputs_history = array(self.database.get_x_history())
        n_samples = inputs_history.shape[0]
        inputs_history = DataConversion.array_to_dict(
            inputs_history, inputs_names, sizes
        )
        for input_name, input_value in inputs_history.items():
            dataset.add_variable(input_name, input_value, input_group)

        def replace_missing_values(
            output_data,  # type: ndarray
            input_data,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Replace the missing output values with NaN.

            Args:
                output_data: The output data with possibly missing values.
                input_data: The input data.

            Returns:
                The output data where missing values have been replaced with NaN.
            """
            if len(inputs_history) != n_samples:
                # There are fewer output values than input values
                n_values = len(input_data)
                output_dimension = output_data.size // n_values
                output_data = output_data.reshape((n_values, output_dimension))
                # Add NaN values at the missing inputs
                # N.B. the inputs are assumed to be in the same order.
                index = 0
                for sub_input_data in input_data:
                    while not array_equal(sub_input_data, inputs_history[index]):
                        output_data = insert(output_data, index, nan, 0)
                        index += 1
                    index += 1
                return insert(output_data, [index] * (n_samples - index), nan, 0)
            else:
                return output_data

        # Add database outputs
        all_data_names = self.database.get_all_data_names()
        outputs_names = list(
            set(all_data_names) - set(inputs_names) - {self.database.ITER_TAG}
        )
        functions_history = []
        for function_name in outputs_names:
            function_history, inputs_history = self.database.get_func_history(
                function_name, x_hist=True
            )
            function_history = replace_missing_values(function_history, inputs_history)
            reshaped_function_history = function_history.reshape((n_samples, -1))
            functions_history.append(reshaped_function_history)
            sizes.update({function_name: functions_history[-1].shape[1]})

        functions_history = concatenate(functions_history, axis=1)
        functions_history = DataConversion.array_to_dict(
            functions_history, outputs_names, sizes
        )
        for output_name, output_value in functions_history.items():
            dataset.add_variable(
                output_name,
                output_value,
                output_group,
                cache_as_input=cache_output_as_input,
            )

        # Add database output gradients
        if export_gradients:
            gradients_history = []
            gradients_names = []

            for function_name in outputs_names:
                if self.database.is_func_grad_history_empty(function_name):
                    continue

                gradient_history, inputs_history = self.database.get_func_grad_history(
                    function_name, x_hist=True
                )
                gradient_history = replace_missing_values(
                    gradient_history, inputs_history
                )
                gradients_history.append(gradient_history.reshape(n_samples, -1))
                gradient_name = Database.get_gradient_name(function_name)
                sizes.update({gradient_name: gradients_history[-1].shape[1]})
                gradients_names.append(Database.get_gradient_name(function_name))

            if gradients_history:
                gradients_history = concatenate(gradients_history, axis=1)
                gradients_history = DataConversion.array_to_dict(
                    gradients_history, gradients_names, sizes
                )
                for gradient_name, gradient_value in gradients_history.items():
                    dataset.add_variable(
                        gradient_name,
                        gradient_value,
                        gradient_group,
                        cache_as_input=cache_output_as_input,
                    )

        return dataset

    @staticmethod
    def __h5_group_to_dict(
        h5_handle,  # type: Union[h5py.File, h5py.Group]
        group_name,  # type: str
    ):  # type: (...) -> Dict[str, Union[str,List[str]]]
        """Convert the values of a hdf5 dataset.

        Values that are of the kind string or bytes are converted
        to string or list of strings.

        Args:
            h5_handle: A hdf5 file or group.
            group_name: The name of the group to be converted.

        Returns:
            The converted dataset.
        """
        converted = {}

        group = get_hdf5_group(h5_handle, group_name)

        for key, value in group.items():
            value = value[()]

            # h5py does not handle bytes natively, it maps it to a numpy generic type
            if isinstance(value, ndarray) and value.dtype.type in (
                numpy.object_,
                numpy.string_,
            ):
                if value.size == 1:
                    value = value[0]
                else:
                    value = value.tolist()

            if isinstance(value, bytes):
                value = value.decode()

            converted[key] = value

        return converted



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Caching module to avoid multiple evaluations of a discipline
************************************************************
"""
from __future__ import division, unicode_literals

import logging
from copy import deepcopy

from gemseo.core.cache import AbstractCache, check_cache_approx, check_cache_equal
from gemseo.utils.data_conversion import DataConversion

LOGGER = logging.getLogger(__name__)


class SimpleCache(AbstractCache):
    """Simple discipline cache based on a dictionary.

    Only caches the last execution.
    """

    def __init__(self, tolerance=0.0, name=None):
        """Initialize cache tolerance. By default, don't use approximate cache. It is up
        to the user to choose to optimize CPU time with this or not.

        could be something like 2 * finfo(float).eps

        Parameters
        ----------
        tolerance : float
            Tolerance that defines if two input vectors
            are equal and cached data shall be returned.
            If 0, no approximation is made. Default: 0.
        name : str
            Name of the cache.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> cache = SimpleCache()
        """
        super(SimpleCache, self).__init__(tolerance, name)
        self.__input_cache = None
        self.__output_cache = None
        self.__jacobian_input_cache = None
        self.__jacobian_cache = None

    def clear(self):
        """Clear the cache.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([.2])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_length()
        1
        >>> cache.clear()
        >>> cache.get_length()
        0
        """
        super(SimpleCache, self).clear()
        self.__input_cache = None
        self.__output_cache = None
        self.__jacobian_input_cache = None
        self.__jacobian_cache = None

    def get_length(self):
        """Get the length of the cache, ie the number of stored elements.

        Returns
        -------
        length : int
            Length of the cache.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_length()
        1
        """
        if not self.__input_cache:
            return 0
        return 1

    @property
    def max_length(self):
        """Get the maximal length of the cache (the maximal number of stored elements).

        Returns
        -------
        length : int
            Maximal length of the cache.
        """
        return 1

    def get_last_cached_inputs(self):
        """Retrieve the last execution inputs.

        Returns
        -------
        inputs : dict
            Last cached inputs.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_last_cached_inputs()
        {'X': array([1.])}
        """
        return self.__input_cache

    def get_last_cached_outputs(self):
        """Retrieve the last execution outputs.

        Returns
        -------
        outputs : dict
            Last cached outputs.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_last_cached_outputs()
        {'y': array([2.])}
        """
        return self.__output_cache

    def get_data(self, index, **options):
        """Returns an elementary sample.

        :param index: sample index.
        :type index: int
        :param options: getter options
        """
        self._check_index(index)
        return {
            self.INPUTS_GROUP: self.__input_cache,
            self.OUTPUTS_GROUP: self.__output_cache,
            self.JACOBIAN_GROUP: self.__jacobian_cache,
        }

    def get_all_data(self, as_iterator=False):
        """Read all the data in the cache.

        Parameters
        ----------
        as_iterator : bool
            If True, return an iterator. Otherwise a dictionary.
            Default: False.


        Returns
        -------
        all_data : dict
            A dictionary of dictionaries for inputs, outputs and jacobian
            where keys are data indices.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_all_data()
        {1: {'inputs': {'x': array([1.])}, 'jacobian': None,
        'outputs': {'y': array([0.2])}}}
        """
        return {1: self.get_data(1)}

    @staticmethod
    def _create_input_cache(input_data, input_names, output_names=None):
        """Create a cache dict for input data.

        :param input_data: the input data to cache
        :param input_names: list of input data names
        :param output_names: list of output data names
        """
        if output_names is None:
            cache_dict = DataConversion.deepcopy_datadict(input_data, input_names)
        else:
            cache_dict = {k: v for k, v in input_data.items() if k in input_names}
            # If also an output, keeps a copy of the original input value
            for key in set(output_names) & set(input_names):
                val = input_data.get(key)
                if val is not None:
                    cache_dict[key] = deepcopy(val)
        return cache_dict

    def cache_outputs(self, input_data, input_names, output_data, output_names=None):
        """Cache data to avoid re evaluation.

        Parameters
        ----------
        input_data : dict
            Input data to cache.
        input_names : list(str)
            List of input data names.
        output_data : dict
            Output data to cache.
        output_names : list(str)
            List of output data names. If None, use all output names.
            Default: None.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache[1]
        {'y': array([2.]), 'x': array([1.])}
        """
        self.__input_cache = self._create_input_cache(
            input_data, input_names, output_names
        )
        self.__output_cache = DataConversion.deepcopy_datadict(
            output_data, output_names
        )

    def get_outputs(self, input_data, input_names=None):
        """Check if the discipline has already been evaluated for the given input data
        dictionary. If True, return the associated cache, otherwise return None.

        Parameters
        ----------
        input_data : dict
            Input data dictionary to test for caching.
        input_names : list(str)
            List of input data names.
            If None, uses them all

        Returns
        -------
        output_data : dict
            Output data if there is no need to evaluate the discipline.
            None otherwise.
        jacobian : dict
            Jacobian if there is no need to evaluate the discipline.
            None otherwise.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_outputs({'x': array([1.])}, ['x'])
        ({'y': array([2.])}, None)
        >>> cache.get_outputs({'x': array([2.])}, ['x'])
        (None, None)
        """
        cached_outs, cached_jac = None, None
        if input_names is None:
            input_names = input_data.keys()
        in_are_cached = self._is_cached(self.__input_cache, input_names, input_data)

        jac_in_are_cached = self._is_cached(
            self.__jacobian_input_cache, input_names, input_data
        )
        if in_are_cached:
            cached_outs = self.__output_cache
        if jac_in_are_cached:
            cached_jac = self.__jacobian_cache

        return cached_outs, cached_jac

    def _is_cached(self, in_cache, input_names, input_data):
        """Check if the input_data dictionary is cached.

        :param in_cache: cached input dictionary
        :param input_names: list of input names
        :param input_data: input dict of data
        :return: True if the data is cached
        """
        if not in_cache or len(in_cache) != len(input_names):
            return False

        if self.tolerance == 0.0:
            if check_cache_equal(input_data, in_cache):
                return True

        else:
            if check_cache_approx(input_data, in_cache, self.tolerance):
                return True
        return False

    def cache_jacobian(self, input_data, input_names, jacobian):
        """Cache jacobian data to avoid re evaluation.

        Parameters
        ----------
        input_data : dict
            Input data to cache.
        input_names : list(str)
            List of input data names.
        jacobian : dict
            Jacobian to cache.

        Examples
        --------
        >>> from gemseo.caches.simple_cache import SimpleCache
        >>> from numpy import array
        >>> cache = SimpleCache()
        >>> data = {'x': array([1.]), 'y': array([2.])}
        >>> jacobian = {'y': {'x': array([3.])}}
        >>> cache.cache_jacobian(data, ['x'], jacobian)
        (None, {'y': {'x': array([3.])}})
        """
        self.__jacobian_input_cache = self._create_input_cache(
            input_data, input_names, jacobian.keys()
        )
        self.__jacobian_cache = jacobian
